"""Tests snowflake storage implementation methods directly."""

import os

import pytest
from ga4gh.vrs import models

from anyvar.storage.snowflake import SnowflakeObjectStore

from .storage_test_funcs import (
    run_alleles_crud,
    run_annotations_crud,
    run_db_lifecycle,
    run_incomplete_objects_error,
    run_mappings_crud,
    run_query_max_rows,
    run_search_alleles,
    run_sequencelocations_crud,
    run_sequencereferences_crud,
)

pytestmark = pytest.mark.snowflake


@pytest.fixture(scope="module")
def snowflake_storage():
    """Reset storage state after each test case"""
    storage = SnowflakeObjectStore(db_url=os.environ.get("ANYVAR_TEST_STORAGE_URI"))
    yield storage
    storage.wipe_db()


@pytest.fixture
def snowflake_storage_with_cleanup(snowflake_storage: SnowflakeObjectStore):
    """Reset storage state after each test case"""
    yield snowflake_storage
    snowflake_storage.wipe_db()


def test_db_lifecycle(
    snowflake_storage: SnowflakeObjectStore,
    validated_vrs_alleles: dict[str, models.Allele],
):
    run_db_lifecycle(snowflake_storage, validated_vrs_alleles)


def test_query_max_rows(
    monkeypatch,
    snowflake_storage_with_cleanup: SnowflakeObjectStore,
    focus_alleles: tuple[models.Allele, models.Allele, models.Allele],
):
    run_query_max_rows(monkeypatch, snowflake_storage_with_cleanup, focus_alleles)


def test_alleles_crud(
    snowflake_storage_with_cleanup: SnowflakeObjectStore,
    focus_alleles: tuple[models.Allele, models.Allele, models.Allele],
    validated_vrs_alleles: dict[str, models.Allele],
):
    run_alleles_crud(
        snowflake_storage_with_cleanup, focus_alleles, validated_vrs_alleles
    )


def test_incomplete_objects_error(snowflake_storage_with_cleanup: SnowflakeObjectStore):
    run_incomplete_objects_error(snowflake_storage_with_cleanup)


def test_sequencelocations_crud(
    snowflake_storage_with_cleanup: SnowflakeObjectStore,
    focus_alleles: tuple[models.Allele, models.Allele, models.Allele],
    validated_vrs_alleles: dict[str, models.Allele],
):
    run_sequencelocations_crud(
        snowflake_storage_with_cleanup, focus_alleles, validated_vrs_alleles
    )


def test_sequencereferences_crud(
    snowflake_storage_with_cleanup: SnowflakeObjectStore,
    focus_alleles: tuple[models.Allele, models.Allele, models.Allele],
    validated_vrs_alleles: dict[str, models.Allele],
):
    run_sequencereferences_crud(
        snowflake_storage_with_cleanup, focus_alleles, validated_vrs_alleles
    )


def test_mappings_crud(
    snowflake_storage_with_cleanup: SnowflakeObjectStore,
    validated_vrs_alleles: dict[str, models.Allele],
):
    run_mappings_crud(snowflake_storage_with_cleanup, validated_vrs_alleles)


def test_annotations_crud(
    snowflake_storage_with_cleanup: SnowflakeObjectStore,
    focus_alleles: tuple[models.Allele, models.Allele, models.Allele],
):
    run_annotations_crud(snowflake_storage_with_cleanup, focus_alleles)


def test_search_alleles(
    snowflake_storage_with_cleanup: SnowflakeObjectStore,
    validated_vrs_alleles: dict[str, models.Allele],
):
    run_search_alleles(snowflake_storage_with_cleanup, validated_vrs_alleles)



"""Provide proxy classes and other tools for translation middleware."""

from .base import (
    TranslationError,
    Translator,
    TranslatorConnectionError,
    TranslatorSetupError,
)

__all__ = [
    "TranslationError",
    "Translator",
    "TranslatorConnectionError",
    "TranslatorSetupError",
]



"""Object mappers for converting between VRS models and database entities."""

from abc import ABC, abstractmethod
from typing import Generic, TypeVar

from ga4gh.vrs import models as vrs_models

from anyvar.core import metadata
from anyvar.storage import orm

A = TypeVar("A")  # Anyvar entity type
D = TypeVar("D")  # DB entity type


class BaseMapper(Generic[A, D], ABC):
    """Base class for all object mappers."""

    @abstractmethod
    def from_db_entity(self, db_entity: D) -> A:
        """Convert DB entity to VRS model."""

    @abstractmethod
    def to_db_entity(self, anyvar_entity: A) -> D:
        """Convert AnyVar entity to DB entity."""


class SequenceReferenceMapper(
    BaseMapper[vrs_models.SequenceReference, orm.SequenceReference]
):
    """Maps between SequenceReference entities."""

    def from_db_entity(
        self, db_entity: orm.SequenceReference
    ) -> vrs_models.SequenceReference:
        """Convert DB SequenceReference to VRS SequenceReference."""
        return vrs_models.SequenceReference(
            type="SequenceReference",
            refgetAccession=db_entity.id,
            moleculeType=db_entity.molecule_type,
        )

    def to_db_entity(
        self, anyvar_entity: vrs_models.SequenceReference
    ) -> orm.SequenceReference:
        """Convert VRS SequenceReference to DB SequenceReference."""
        return orm.SequenceReference(
            id=anyvar_entity.refgetAccession,  # Use refgetAccession as primary key
            molecule_type=anyvar_entity.moleculeType,
        )


class SequenceLocationMapper(BaseMapper[vrs_models.SequenceLocation, orm.Location]):
    """Maps between SequenceLocation entities."""

    def __init__(self) -> None:
        """Initialize SequenceLocationMapper."""
        self.seq_ref_mapper = SequenceReferenceMapper()

    def from_db_entity(self, db_entity: orm.Location) -> vrs_models.SequenceLocation:
        """Convert DB Location to VRS SequenceLocation."""
        # Handle range vs simple coordinates
        start = self._resolve_coordinate_from_db(
            simple=db_entity.start,
            start=db_entity.start_outer,
            end=db_entity.start_inner,
        )
        end = self._resolve_coordinate_from_db(
            simple=db_entity.end, start=db_entity.end_outer, end=db_entity.end_inner
        )

        return vrs_models.SequenceLocation(
            id=db_entity.id,
            digest=db_entity.id.removeprefix("ga4gh:SL."),
            type="SequenceLocation",
            sequenceReference=self.seq_ref_mapper.from_db_entity(
                db_entity.sequence_reference
            ),
            start=start,
            end=end,
        )

    def to_db_entity(self, anyvar_entity: vrs_models.SequenceLocation) -> orm.Location:
        """Convert VRS SequenceLocation to DB Location.

        :raise AttributeError: if `.id` field is missing
        """
        if not anyvar_entity.id:
            msg = "`.id` property is required for storing in DB"
            raise AttributeError(msg)

        # Convert VRS int/Range coordinates to DB fields
        start_simple, start_outer, start_inner = self._resolve_coordinate_to_db(
            anyvar_entity.start
        )
        end_simple, end_outer, end_inner = self._resolve_coordinate_to_db(
            anyvar_entity.end
        )

        # Construct orm.Location and delegate to orm.SequenceReference mapper
        return orm.Location(
            id=anyvar_entity.id,
            sequence_reference_id=anyvar_entity.sequenceReference.refgetAccession,
            sequence_reference=self.seq_ref_mapper.to_db_entity(
                anyvar_entity.sequenceReference
            ),
            start=start_simple,
            end=end_simple,
            start_outer=start_outer,
            start_inner=start_inner,
            end_outer=end_outer,
            end_inner=end_inner,
        )

    def _resolve_coordinate_from_db(
        self, simple: int | None, start: int | None, end: int | None
    ) -> int | vrs_models.Range | None:
        """Resolve coordinate from DB fields to VRS format."""
        if simple is not None:
            return simple
        if start is not None and end is not None:
            return vrs_models.Range([start, end])
        return None

    def _resolve_coordinate_to_db(
        self, coordinate: int | vrs_models.Range | None
    ) -> tuple[int | None, int | None, int | None]:
        """Resolve VRS coordinate to DB fields."""
        if coordinate is None:
            return None, None, None
        if isinstance(coordinate, vrs_models.Range):
            return None, coordinate.root[0], coordinate.root[1]
        if isinstance(coordinate, int):
            return coordinate, None, None
        return None, None, None


class AlleleMapper(BaseMapper[vrs_models.Allele, orm.Allele]):
    """Maps between Allele entities."""

    def __init__(self) -> None:
        """Initialize AlleleMapper."""
        self.location_mapper = SequenceLocationMapper()

    def from_db_entity(self, db_entity: orm.Allele) -> vrs_models.Allele:
        """Convert DB Allele to VRS Allele."""
        # Reconstruct state from JSONB
        state = self._reconstruct_state(db_entity.state)

        # Construct orm.Allele and delegate to orm.Location mapper
        return vrs_models.Allele(
            id=db_entity.id,
            digest=db_entity.id.removeprefix("ga4gh:VA."),
            type="Allele",
            location=self.location_mapper.from_db_entity(db_entity.location),
            state=state,
        )

    def to_db_entity(self, anyvar_entity: vrs_models.Allele) -> orm.Allele:
        """Convert VRS Allele to DB Allele.

        :raise AttributeError: if `.id` field is missing, or uses
            IRI references in place of full SequenceLocation/SequenceReference
        """
        if not anyvar_entity.id:
            msg = "`.id` property is required for storing in DB"
            raise AttributeError(msg)

        # Serialize state
        state_dict = anyvar_entity.state.model_dump(exclude_none=True)

        # purposely trip attribute/type errors below
        return orm.Allele(
            id=anyvar_entity.id,
            location_id=anyvar_entity.location.id,  # type: ignore[reportAttributeAccessIssue]
            location=self.location_mapper.to_db_entity(anyvar_entity.location),  # type: ignore[reportArgumentType]
            state=state_dict,
        )

    def _reconstruct_state(
        self, state_data: dict
    ) -> (
        vrs_models.LiteralSequenceExpression
        | vrs_models.ReferenceLengthExpression
        | vrs_models.LengthExpression
    ):
        """Reconstruct state object from JSONB data."""
        state_type = state_data.get("type")

        if state_type == "LiteralSequenceExpression":
            return vrs_models.LiteralSequenceExpression(**state_data)
        if state_type == "ReferenceLengthExpression":
            return vrs_models.ReferenceLengthExpression(**state_data)
        if state_type == "LengthExpression":
            return vrs_models.LengthExpression(**state_data)
        raise ValueError(f"Unknown state type '{state_type}' from: {state_data}")


class VariationMappingMapper(
    BaseMapper[metadata.VariationMapping, orm.VariationMapping]
):
    """Maps between VariationMapping entities."""

    def from_db_entity(
        self, db_entity: orm.VariationMapping
    ) -> metadata.VariationMapping:
        """Convert DB instance into business logic object"""
        mapping_type = metadata.VariationMappingType(db_entity.mapping_type)
        return metadata.VariationMapping(
            source_id=db_entity.source_id,
            dest_id=db_entity.dest_id,
            mapping_type=mapping_type,
        )

    def to_db_entity(
        self, anyvar_entity: metadata.VariationMapping
    ) -> orm.VariationMapping:
        """Convert VariationMapping object to DB mapping instance."""
        return orm.VariationMapping(
            source_id=anyvar_entity.source_id,
            dest_id=anyvar_entity.dest_id,
            mapping_type=anyvar_entity.mapping_type,
        )


class AnnotationMapper(BaseMapper[metadata.Annotation, orm.Annotation]):
    """Maps between Annotations entities."""

    def from_db_entity(self, db_entity: orm.Annotation) -> metadata.Annotation:
        """Convert DB Annotation to AnyVar Annotation.

        :param db_entity: An ORM Annotation instance
        :return: An Anyvar Annotation instance
        """
        return metadata.Annotation(
            object_id=db_entity.object_id,
            annotation_type=db_entity.annotation_type,
            annotation_value=db_entity.annotation_value,
        )

    def to_db_entity(self, anyvar_entity: metadata.Annotation) -> orm.Annotation:
        """Convert AnyVar Annotation into DB Annotation.

        :param anyvar_entity: An Anyvar Annotation instance
        :return: An ORM model Annotation instance
        """
        return orm.Annotation(**anyvar_entity.model_dump())



import os
from unittest.mock import MagicMock, Mock, patch

import jwt
import pytest
from fastapi import HTTPException

from anyvar.restapi.auth import (
    BearerTokenAuthConfig,
    ValidTokenCache,
    get_token_auth_dependency,
    is_jwt_token_valid,
)


@pytest.mark.asyncio
async def test_no_auth():
    """Test that no authentication is required when no auth config is set."""
    dependency = get_token_auth_dependency()
    assert callable(dependency)
    assert dependency.__name__ == "no_auth_required"
    assert (await dependency()) is None


def test_config_default_initialization():
    """Test default initialization of BearerTokenAuthConfig."""
    config = BearerTokenAuthConfig()
    assert config.token_list == []
    assert config.issuer_url is None
    assert config.audiences == []
    assert config.appids == []
    assert config.scopes == []
    assert config.emails == []
    assert config.subjects == []
    assert config.jwks_uri is None


def test_config_initialization_with_values():
    """Test initialization with explicit values."""
    config = BearerTokenAuthConfig(
        token_list=["token1", "token2"],
        issuer_url="https://issuer.example.com",
        audiences=["https://api.example.com"],
        appids=["app-id-123"],
        scopes=["read", "write"],
        emails=["user@example.com"],
        subjects=["sub123"],
        jwks_uri="https://issuer.example.com/.well-known/jwks.json",
    )
    assert config.token_list == ["token1", "token2"]
    assert config.issuer_url == "https://issuer.example.com"
    assert config.audiences == ["https://api.example.com"]
    assert config.appids == ["app-id-123"]
    assert config.scopes == ["read", "write"]
    assert config.emails == ["user@example.com"]
    assert config.subjects == ["sub123"]
    assert config.jwks_uri == "https://issuer.example.com/.well-known/jwks.json"


@patch.dict(os.environ, {}, clear=True)
def test_config_from_env_empty():
    """Test from_env with no environment variables set."""
    config = BearerTokenAuthConfig.from_env()
    assert config.token_list == []
    assert config.issuer_url is None
    assert config.audiences == []
    assert config.appids == []
    assert config.scopes == []
    assert config.emails == []
    assert config.subjects == []
    assert config.jwks_uri is None


@patch.dict(
    os.environ,
    {
        "ANYVAR_AUTH_TOKEN_LIST": "token1,token2,token3",
        "ANYVAR_AUTH_ISSUER_URL": "https://issuer.example.com",
        "ANYVAR_AUTH_AUDIENCES": "https://api.example.com,api://my-api",
        "ANYVAR_AUTH_APPIDS": "app-id-123,app-id-456",
        "ANYVAR_AUTH_SCOPES": "read,write",
        "ANYVAR_AUTH_EMAILS": "User@Example.Com,admin@example.com",
        "ANYVAR_AUTH_SUBJECTS": "sub123,sub456",
        "ANYVAR_AUTH_JWKS_URI": "https://issuer.example.com/.well-known/jwks.json",
    },
)
def test_config_from_env_with_values():
    """Test from_env with environment variables set."""
    config = BearerTokenAuthConfig.from_env()
    assert config.token_list == ["token1", "token2", "token3"]
    assert config.issuer_url == "https://issuer.example.com"
    assert config.audiences == ["https://api.example.com", "api://my-api"]
    assert config.appids == ["app-id-123", "app-id-456"]
    assert config.scopes == ["read", "write"]
    assert config.emails == ["user@example.com", "admin@example.com"]
    assert config.subjects == ["sub123", "sub456"]
    assert config.jwks_uri == "https://issuer.example.com/.well-known/jwks.json"


@patch.dict(
    os.environ,
    {"ANYVAR_AUTH_TOKEN_LIST": " token1 , token2 , , token3 "},
)
def test_config_from_env_strips_whitespace():
    """Test that from_env strips whitespace and empty values."""
    config = BearerTokenAuthConfig.from_env()
    assert config.token_list == ["token1", "token2", "token3"]


def test_config_get_signing_key_from_jwt_no_jwks_uri():
    """Test get_signing_key_from_jwt raises error when JWKS URI not configured."""
    config = BearerTokenAuthConfig()
    with pytest.raises(ValueError, match="JWKS URI is not configured"):
        config.get_signing_key_from_jwt("fake_token")


@patch("anyvar.restapi.auth.PyJWKClient")
def test_config_get_signing_key_from_jwt(mock_jwk_client_class):
    """Test get_signing_key_from_jwt successfully retrieves signing key."""
    mock_client = MagicMock()
    mock_signing_key = Mock()
    mock_client.get_signing_key_from_jwt.return_value = mock_signing_key
    mock_jwk_client_class.return_value = mock_client

    config = BearerTokenAuthConfig(jwks_uri="https://example.com/jwks")
    result = config.get_signing_key_from_jwt("fake_token")

    assert result == mock_signing_key
    mock_jwk_client_class.assert_called_once_with(
        "https://example.com/jwks", cache_keys=True, lifespan=3600
    )
    mock_client.get_signing_key_from_jwt.assert_called_once_with("fake_token")


@patch("anyvar.restapi.auth.PyJWKClient")
def test_config_get_signing_key_from_jwt_caches_client(mock_jwk_client_class):
    """Test that PyJWKClient is cached after first call."""
    mock_client = MagicMock()
    mock_jwk_client_class.return_value = mock_client

    config = BearerTokenAuthConfig(jwks_uri="https://example.com/jwks")
    config.get_signing_key_from_jwt("token1")
    config.get_signing_key_from_jwt("token2")

    # Should only create client once
    assert mock_jwk_client_class.call_count == 1


def test_valid_token_cache_initialization():
    """Test ValidTokenCache initialization."""
    cache = ValidTokenCache()
    assert cache.cache == {}
    assert isinstance(cache.last_purge, float)


def test_valid_token_cache_is_token_valid_not_in_cache():
    """Test is_token_valid returns False when token not in cache."""
    cache = ValidTokenCache()
    assert cache.is_token_valid("nonexistent_token") is False


def test_valid_token_cache_is_token_valid_expired():
    """Test is_token_valid returns False and removes expired token."""
    cache = ValidTokenCache()
    expired_time = 1000.0
    cache.cache["expired_token"] = expired_time

    with patch("time.time", return_value=2000.0):
        assert cache.is_token_valid("expired_token") is False

    # Token should be removed from cache
    assert "expired_token" not in cache.cache


def test_valid_token_cache_is_token_valid_not_expired():
    """Test is_token_valid returns True for valid token."""
    cache = ValidTokenCache()
    future_time = 9999999999.0
    cache.cache["valid_token"] = future_time

    with patch("time.time", return_value=1000.0):
        assert cache.is_token_valid("valid_token") is True

    # Token should remain in cache
    assert "valid_token" in cache.cache


def test_valid_token_cache_add_token():
    """Test add_token adds token with expiry to cache."""
    cache = ValidTokenCache()
    cache.add_token("new_token", 5000)

    assert cache.cache["new_token"] == 5000


def test_valid_token_cache_purge_no_purge_before_hour():
    """Test purge does not run if called before 1 hour."""
    cache = ValidTokenCache()
    cache.last_purge = 1000.0
    cache.cache["token1"] = 500.0  # Expired
    cache.cache["token2"] = 9999.0  # Valid

    with patch("time.time", return_value=2000.0):  # Less than 3600 seconds later
        cache.purge()

    # No purging should happen
    assert "token1" in cache.cache
    assert cache.last_purge == 1000.0


def test_valid_token_cache_purge_removes_expired():
    """Test purge removes expired tokens after 1 hour."""
    cache = ValidTokenCache()
    cache.last_purge = 1000.0
    cache.cache["expired_token"] = 3000.0
    cache.cache["valid_token"] = 9999.0

    with patch("time.time", return_value=5000.0):  # More than 3600 seconds later
        cache.purge()

    assert "expired_token" not in cache.cache
    assert "valid_token" in cache.cache
    assert cache.last_purge == 5000.0


def test_valid_token_cache_purge_updates_last_purge_time():
    """Test purge updates last_purge timestamp."""
    cache = ValidTokenCache()
    cache.last_purge = 1000.0

    with patch("time.time", return_value=5000.0):
        cache.purge()

    assert cache.last_purge == 5000.0


# Tests for is_jwt_token_valid function


def test_is_jwt_token_valid_no_jwks_uri():
    """Test is_jwt_token_valid returns False when JWKS URI not configured."""
    config = BearerTokenAuthConfig(issuer_url="https://issuer.example.com")
    cache = ValidTokenCache()

    result = is_jwt_token_valid(config, "fake_token", cache)

    assert result is False


def test_is_jwt_token_valid_no_issuer_url():
    """Test is_jwt_token_valid returns False when issuer URL not configured."""
    config = BearerTokenAuthConfig(jwks_uri="https://example.com/jwks")
    cache = ValidTokenCache()

    result = is_jwt_token_valid(config, "fake_token", cache)

    assert result is False


def test_is_jwt_token_valid_cached_token():
    """Test is_jwt_token_valid returns True for cached valid token."""
    config = BearerTokenAuthConfig(
        issuer_url="https://issuer.example.com", jwks_uri="https://example.com/jwks"
    )
    cache = ValidTokenCache()
    cache.add_token("cached_token", 9999999999.0)

    result = is_jwt_token_valid(config, "cached_token", cache)

    assert result is True


@patch("anyvar.restapi.auth.jwt.decode")
def test_is_jwt_token_valid_successful_validation(mock_jwt_decode):
    """Test is_jwt_token_valid with successful JWT validation."""
    config = BearerTokenAuthConfig(
        issuer_url="https://issuer.example.com", jwks_uri="https://example.com/jwks"
    )
    cache = ValidTokenCache()

    # Mock the JWT decode to return valid claims
    mock_jwt_decode.return_value = {
        "iss": "https://issuer.example.com",
        "exp": 9999999999,
        "iat": 1000000000,
        "sub": "user123",
    }

    # Mock get_signing_key_from_jwt
    with patch.object(config, "get_signing_key_from_jwt") as mock_get_key:
        mock_signing_key = Mock()
        mock_signing_key.key = "fake_key"
        mock_get_key.return_value = mock_signing_key

        result = is_jwt_token_valid(config, "valid_token", cache)

    assert result is True
    mock_jwt_decode.assert_called_once()


@patch("anyvar.restapi.auth.jwt.decode")
def test_is_jwt_token_valid_wrong_issuer(mock_jwt_decode):
    """Test is_jwt_token_valid returns False when issuer doesn't match."""
    config = BearerTokenAuthConfig(
        issuer_url="https://issuer.example.com", jwks_uri="https://example.com/jwks"
    )
    cache = ValidTokenCache()

    mock_jwt_decode.return_value = {
        "iss": "https://wrong-issuer.example.com",
        "exp": 9999999999,
        "iat": 1000000000,
        "sub": "user123",
    }

    with patch.object(config, "get_signing_key_from_jwt") as mock_get_key:
        mock_signing_key = Mock()
        mock_signing_key.key = "fake_key"
        mock_get_key.return_value = mock_signing_key

        result = is_jwt_token_valid(config, "token", cache)

    assert result is False


@patch("anyvar.restapi.auth.jwt.decode")
def test_is_jwt_token_valid_with_valid_audience_string(mock_jwt_decode):
    """Test is_jwt_token_valid with valid audience as string."""
    config = BearerTokenAuthConfig(
        issuer_url="https://issuer.example.com",
        jwks_uri="https://example.com/jwks",
        audiences=["https://api.example.com"],
    )
    cache = ValidTokenCache()

    mock_jwt_decode.return_value = {
        "iss": "https://issuer.example.com",
        "aud": "https://api.example.com",
        "exp": 9999999999,
        "iat": 1000000000,
        "sub": "user123",
    }

    with patch.object(config, "get_signing_key_from_jwt") as mock_get_key:
        mock_signing_key = Mock()
        mock_signing_key.key = "fake_key"
        mock_get_key.return_value = mock_signing_key

        result = is_jwt_token_valid(config, "token", cache)

    assert result is True


@patch("anyvar.restapi.auth.jwt.decode")
def test_is_jwt_token_valid_with_valid_audience_list(mock_jwt_decode):
    """Test is_jwt_token_valid with valid audience as list."""
    config = BearerTokenAuthConfig(
        issuer_url="https://issuer.example.com",
        jwks_uri="https://example.com/jwks",
        audiences=["https://api.example.com", "api://my-api"],
    )
    cache = ValidTokenCache()

    mock_jwt_decode.return_value = {
        "iss": "https://issuer.example.com",
        "aud": ["https://api.example.com", "https://other-api.example.com"],
        "exp": 9999999999,
        "iat": 1000000000,
        "sub": "user123",
    }

    with patch.object(config, "get_signing_key_from_jwt") as mock_get_key:
        mock_signing_key = Mock()
        mock_signing_key.key = "fake_key"
        mock_get_key.return_value = mock_signing_key

        result = is_jwt_token_valid(config, "token", cache)

    assert result is True


@patch("anyvar.restapi.auth.jwt.decode")
def test_is_jwt_token_valid_with_invalid_audience(mock_jwt_decode):
    """Test is_jwt_token_valid returns False when audience doesn't match."""
    config = BearerTokenAuthConfig(
        issuer_url="https://issuer.example.com",
        jwks_uri="https://example.com/jwks",
        audiences=["https://api.example.com"],
    )
    cache = ValidTokenCache()

    mock_jwt_decode.return_value = {
        "iss": "https://issuer.example.com",
        "aud": "https://wrong-api.example.com",
        "exp": 9999999999,
        "iat": 1000000000,
        "sub": "user123",
    }

    with patch.object(config, "get_signing_key_from_jwt") as mock_get_key:
        mock_signing_key = Mock()
        mock_signing_key.key = "fake_key"
        mock_get_key.return_value = mock_signing_key

        result = is_jwt_token_valid(config, "token", cache)

    assert result is False


@patch("anyvar.restapi.auth.jwt.decode")
def test_is_jwt_token_valid_with_valid_appid(mock_jwt_decode):
    """Test is_jwt_token_valid with valid appid claim."""
    config = BearerTokenAuthConfig(
        issuer_url="https://issuer.example.com",
        jwks_uri="https://example.com/jwks",
        appids=["app-id-123", "app-id-456"],
    )
    cache = ValidTokenCache()

    mock_jwt_decode.return_value = {
        "iss": "https://issuer.example.com",
        "appid": "app-id-123",
        "exp": 9999999999,
        "iat": 1000000000,
        "sub": "user123",
    }

    with patch.object(config, "get_signing_key_from_jwt") as mock_get_key:
        mock_signing_key = Mock()
        mock_signing_key.key = "fake_key"
        mock_get_key.return_value = mock_signing_key

        result = is_jwt_token_valid(config, "token", cache)

    assert result is True


@patch("anyvar.restapi.auth.jwt.decode")
def test_is_jwt_token_valid_with_valid_azp(mock_jwt_decode):
    """Test is_jwt_token_valid with valid azp claim (fallback for appid)."""
    config = BearerTokenAuthConfig(
        issuer_url="https://issuer.example.com",
        jwks_uri="https://example.com/jwks",
        appids=["app-id-123"],
    )
    cache = ValidTokenCache()

    mock_jwt_decode.return_value = {
        "iss": "https://issuer.example.com",
        "appid": "app-id-123",
        "exp": 9999999999,
        "iat": 1000000000,
        "sub": "user123",
    }

    with patch.object(config, "get_signing_key_from_jwt") as mock_get_key:
        mock_signing_key = Mock()
        mock_signing_key.key = "fake_key"
        mock_get_key.return_value = mock_signing_key

        result = is_jwt_token_valid(config, "token", cache)

    assert result is True


@patch("anyvar.restapi.auth.jwt.decode")
def test_is_jwt_token_valid_with_invalid_appid(mock_jwt_decode):
    """Test is_jwt_token_valid returns False when appid doesn't match."""
    config = BearerTokenAuthConfig(
        issuer_url="https://issuer.example.com",
        jwks_uri="https://example.com/jwks",
        appids=["app-id-123"],
    )
    cache = ValidTokenCache()

    mock_jwt_decode.return_value = {
        "iss": "https://issuer.example.com",
        "appid": "wrong-app-id",
        "exp": 9999999999,
        "iat": 1000000000,
        "sub": "user123",
    }

    with patch.object(config, "get_signing_key_from_jwt") as mock_get_key:
        mock_signing_key = Mock()
        mock_signing_key.key = "fake_key"
        mock_get_key.return_value = mock_signing_key

        result = is_jwt_token_valid(config, "token", cache)

    assert result is False


@patch("anyvar.restapi.auth.jwt.decode")
def test_is_jwt_token_valid_with_valid_scopes(mock_jwt_decode):
    """Test is_jwt_token_valid with valid scopes."""
    config = BearerTokenAuthConfig(
        issuer_url="https://issuer.example.com",
        jwks_uri="https://example.com/jwks",
        scopes=["read", "write"],
    )
    cache = ValidTokenCache()

    mock_jwt_decode.return_value = {
        "iss": "https://issuer.example.com",
        "exp": 9999999999,
        "iat": 1000000000,
        "sub": "user123",
        "scope": "read write execute",
    }

    with patch.object(config, "get_signing_key_from_jwt") as mock_get_key:
        mock_signing_key = Mock()
        mock_signing_key.key = "fake_key"
        mock_get_key.return_value = mock_signing_key

        result = is_jwt_token_valid(config, "token", cache)

    assert result is True


@patch("anyvar.restapi.auth.jwt.decode")
def test_is_jwt_token_valid_with_scope_as_list(mock_jwt_decode):
    """Test is_jwt_token_valid with scope as list."""
    config = BearerTokenAuthConfig(
        issuer_url="https://issuer.example.com",
        jwks_uri="https://example.com/jwks",
        scopes=["read"],
    )
    cache = ValidTokenCache()

    mock_jwt_decode.return_value = {
        "iss": "https://issuer.example.com",
        "exp": 9999999999,
        "iat": 1000000000,
        "sub": "user123",
        "scope": ["read", "write"],
    }

    with patch.object(config, "get_signing_key_from_jwt") as mock_get_key:
        mock_signing_key = Mock()
        mock_signing_key.key = "fake_key"
        mock_get_key.return_value = mock_signing_key

        result = is_jwt_token_valid(config, "token", cache)

    assert result is True


@patch("anyvar.restapi.auth.jwt.decode")
def test_is_jwt_token_valid_with_invalid_scopes(mock_jwt_decode):
    """Test is_jwt_token_valid returns False when required scopes not present."""
    config = BearerTokenAuthConfig(
        issuer_url="https://issuer.example.com",
        jwks_uri="https://example.com/jwks",
        scopes=["admin"],
    )
    cache = ValidTokenCache()

    mock_jwt_decode.return_value = {
        "iss": "https://issuer.example.com",
        "exp": 9999999999,
        "iat": 1000000000,
        "sub": "user123",
        "scope": "read write",
    }

    with patch.object(config, "get_signing_key_from_jwt") as mock_get_key:
        mock_signing_key = Mock()
        mock_signing_key.key = "fake_key"
        mock_get_key.return_value = mock_signing_key

        result = is_jwt_token_valid(config, "token", cache)

    assert result is False


@patch("anyvar.restapi.auth.jwt.decode")
def test_is_jwt_token_valid_with_valid_email(mock_jwt_decode):
    """Test is_jwt_token_valid with valid email."""
    config = BearerTokenAuthConfig(
        issuer_url="https://issuer.example.com",
        jwks_uri="https://example.com/jwks",
        emails=["user@example.com"],
    )
    cache = ValidTokenCache()

    mock_jwt_decode.return_value = {
        "iss": "https://issuer.example.com",
        "exp": 9999999999,
        "iat": 1000000000,
        "sub": "user123",
        "email": "user@example.com",
    }

    with patch.object(config, "get_signing_key_from_jwt") as mock_get_key:
        mock_signing_key = Mock()
        mock_signing_key.key = "fake_key"
        mock_get_key.return_value = mock_signing_key

        result = is_jwt_token_valid(config, "token", cache)

    assert result is True


@patch("anyvar.restapi.auth.jwt.decode")
def test_is_jwt_token_valid_with_case_insensitive_email(mock_jwt_decode):
    """Test is_jwt_token_valid handles email case insensitivity."""
    config = BearerTokenAuthConfig(
        issuer_url="https://issuer.example.com",
        jwks_uri="https://example.com/jwks",
        emails=["user@example.com"],
    )
    cache = ValidTokenCache()

    mock_jwt_decode.return_value = {
        "iss": "https://issuer.example.com",
        "exp": 9999999999,
        "iat": 1000000000,
        "sub": "user123",
        "email": "User@Example.Com",
    }

    with patch.object(config, "get_signing_key_from_jwt") as mock_get_key:
        mock_signing_key = Mock()
        mock_signing_key.key = "fake_key"
        mock_get_key.return_value = mock_signing_key

        result = is_jwt_token_valid(config, "token", cache)

    assert result is True


@patch("anyvar.restapi.auth.jwt.decode")
def test_is_jwt_token_valid_with_invalid_email(mock_jwt_decode):
    """Test is_jwt_token_valid returns False when email not in allowed list."""
    config = BearerTokenAuthConfig(
        issuer_url="https://issuer.example.com",
        jwks_uri="https://example.com/jwks",
        emails=["allowed@example.com"],
    )
    cache = ValidTokenCache()

    mock_jwt_decode.return_value = {
        "iss": "https://issuer.example.com",
        "exp": 9999999999,
        "iat": 1000000000,
        "sub": "user123",
        "email": "notallowed@example.com",
    }

    with patch.object(config, "get_signing_key_from_jwt") as mock_get_key:
        mock_signing_key = Mock()
        mock_signing_key.key = "fake_key"
        mock_get_key.return_value = mock_signing_key

        result = is_jwt_token_valid(config, "token", cache)

    assert result is False


@patch("anyvar.restapi.auth.jwt.decode")
def test_is_jwt_token_valid_with_valid_subject(mock_jwt_decode):
    """Test is_jwt_token_valid with valid subject."""
    config = BearerTokenAuthConfig(
        issuer_url="https://issuer.example.com",
        jwks_uri="https://example.com/jwks",
        subjects=["sub123", "sub456"],
    )
    cache = ValidTokenCache()

    mock_jwt_decode.return_value = {
        "iss": "https://issuer.example.com",
        "exp": 9999999999,
        "iat": 1000000000,
        "sub": "sub123",
    }

    with patch.object(config, "get_signing_key_from_jwt") as mock_get_key:
        mock_signing_key = Mock()
        mock_signing_key.key = "fake_key"
        mock_get_key.return_value = mock_signing_key

        result = is_jwt_token_valid(config, "token", cache)

    assert result is True


@patch("anyvar.restapi.auth.jwt.decode")
def test_is_jwt_token_valid_with_invalid_subject(mock_jwt_decode):
    """Test is_jwt_token_valid returns False when subject not in allowed list."""
    config = BearerTokenAuthConfig(
        issuer_url="https://issuer.example.com",
        jwks_uri="https://example.com/jwks",
        subjects=["sub123"],
    )
    cache = ValidTokenCache()

    mock_jwt_decode.return_value = {
        "iss": "https://issuer.example.com",
        "exp": 9999999999,
        "iat": 1000000000,
        "sub": "sub999",
    }

    with patch.object(config, "get_signing_key_from_jwt") as mock_get_key:
        mock_signing_key = Mock()
        mock_signing_key.key = "fake_key"
        mock_get_key.return_value = mock_signing_key

        result = is_jwt_token_valid(config, "token", cache)

    assert result is False


@patch("anyvar.restapi.auth.jwt.decode")
def test_is_jwt_token_valid_with_all_validations(mock_jwt_decode):
    """Test is_jwt_token_valid with all validation criteria."""
    config = BearerTokenAuthConfig(
        issuer_url="https://issuer.example.com",
        jwks_uri="https://example.com/jwks",
        scopes=["read"],
        emails=["user@example.com"],
        subjects=["sub123"],
    )
    cache = ValidTokenCache()

    mock_jwt_decode.return_value = {
        "iss": "https://issuer.example.com",
        "exp": 9999999999,
        "iat": 1000000000,
        "scope": "read write",
        "email": "user@example.com",
        "sub": "sub123",
    }

    with patch.object(config, "get_signing_key_from_jwt") as mock_get_key:
        mock_signing_key = Mock()
        mock_signing_key.key = "fake_key"
        mock_get_key.return_value = mock_signing_key

        result = is_jwt_token_valid(config, "token", cache)

    assert result is True


@patch("anyvar.restapi.auth.jwt.decode")
def test_is_jwt_token_valid_jwt_decode_exception(mock_jwt_decode):
    """Test is_jwt_token_valid returns False when JWT decode raises exception."""
    config = BearerTokenAuthConfig(
        issuer_url="https://issuer.example.com", jwks_uri="https://example.com/jwks"
    )
    cache = ValidTokenCache()

    mock_jwt_decode.side_effect = jwt.DecodeError("Invalid token")

    with patch.object(config, "get_signing_key_from_jwt") as mock_get_key:
        mock_signing_key = Mock()
        mock_signing_key.key = "fake_key"
        mock_get_key.return_value = mock_signing_key

        result = is_jwt_token_valid(config, "invalid_token", cache)

    assert result is False


@patch("anyvar.restapi.auth.jwt.decode")
def test_is_jwt_token_valid_expired_token(mock_jwt_decode):
    """Test is_jwt_token_valid returns False when token is expired."""
    config = BearerTokenAuthConfig(
        issuer_url="https://issuer.example.com", jwks_uri="https://example.com/jwks"
    )
    cache = ValidTokenCache()

    mock_jwt_decode.side_effect = jwt.ExpiredSignatureError("Token expired")

    with patch.object(config, "get_signing_key_from_jwt") as mock_get_key:
        mock_signing_key = Mock()
        mock_signing_key.key = "fake_key"
        mock_get_key.return_value = mock_signing_key

        result = is_jwt_token_valid(config, "expired_token", cache)

    assert result is False


def test_is_jwt_token_valid_get_signing_key_exception():
    """Test is_jwt_token_valid returns False when get_signing_key raises exception."""
    config = BearerTokenAuthConfig(
        issuer_url="https://issuer.example.com", jwks_uri="https://example.com/jwks"
    )
    cache = ValidTokenCache()

    with patch.object(
        config, "get_signing_key_from_jwt", side_effect=Exception("Key error")
    ):
        result = is_jwt_token_valid(config, "token", cache)

    assert result is False


# Tests for check_bearer_token function (via get_token_auth_dependency)


@pytest.mark.asyncio
@patch.dict(
    os.environ,
    {"ANYVAR_AUTH_TOKEN_LIST": "valid_token_1,valid_token_2"},
)
async def test_check_bearer_token_with_valid_literal_token():
    """Test check_bearer_token accepts valid literal token."""
    dependency = get_token_auth_dependency()

    # Create a mock HTTPBearer credentials object
    mock_auth = Mock()
    mock_auth.credentials = "valid_token_1"

    # Should not raise exception
    result = await dependency(mock_auth)
    assert result is None


@pytest.mark.asyncio
@patch.dict(
    os.environ,
    {"ANYVAR_AUTH_TOKEN_LIST": "valid_token_1,valid_token_2"},
)
async def test_check_bearer_token_with_invalid_literal_token():
    """Test check_bearer_token rejects invalid literal token."""
    dependency = get_token_auth_dependency()

    mock_auth = Mock()
    mock_auth.credentials = "invalid_token"

    with pytest.raises(HTTPException) as exc_info:
        await dependency(mock_auth)

    assert exc_info.value.status_code == 401
    assert exc_info.value.detail == "Invalid token"


@pytest.mark.asyncio
@patch.dict(
    os.environ,
    {
        "ANYVAR_AUTH_ISSUER_URL": "https://issuer.example.com",
        "ANYVAR_AUTH_JWKS_URI": "https://example.com/jwks",
    },
)
@patch("anyvar.restapi.auth.is_jwt_token_valid")
async def test_check_bearer_token_with_valid_jwt(mock_is_jwt_valid):
    """Test check_bearer_token accepts valid JWT token."""
    mock_is_jwt_valid.return_value = True

    dependency = get_token_auth_dependency()

    mock_auth = Mock()
    mock_auth.credentials = "valid.jwt.token"

    result = await dependency(mock_auth)
    assert result is None
    mock_is_jwt_valid.assert_called_once()


@pytest.mark.asyncio
@patch.dict(
    os.environ,
    {
        "ANYVAR_AUTH_ISSUER_URL": "https://issuer.example.com",
        "ANYVAR_AUTH_JWKS_URI": "https://example.com/jwks",
    },
)
@patch("anyvar.restapi.auth.is_jwt_token_valid")
async def test_check_bearer_token_with_invalid_jwt(mock_is_jwt_valid):
    """Test check_bearer_token rejects invalid JWT token."""
    mock_is_jwt_valid.return_value = False

    dependency = get_token_auth_dependency()

    mock_auth = Mock()
    mock_auth.credentials = "invalid.jwt.token"

    with pytest.raises(HTTPException) as exc_info:
        await dependency(mock_auth)

    assert exc_info.value.status_code == 401
    assert exc_info.value.detail == "Invalid token"


@pytest.mark.asyncio
@patch.dict(
    os.environ,
    {
        "ANYVAR_AUTH_TOKEN_LIST": "literal_token",
        "ANYVAR_AUTH_ISSUER_URL": "https://issuer.example.com",
        "ANYVAR_AUTH_JWKS_URI": "https://example.com/jwks",
    },
)
async def test_check_bearer_token_literal_token_takes_precedence():
    """Test check_bearer_token checks literal tokens first."""
    dependency = get_token_auth_dependency()

    mock_auth = Mock()
    mock_auth.credentials = "literal_token"

    # Should succeed without calling JWT validation
    with patch("anyvar.restapi.auth.is_jwt_token_valid") as mock_is_jwt_valid:
        result = await dependency(mock_auth)
        assert result is None
        # JWT validation should not be called since literal token matched
        mock_is_jwt_valid.assert_not_called()


@pytest.mark.asyncio
@patch.dict(
    os.environ,
    {
        "ANYVAR_AUTH_TOKEN_LIST": "literal_token",
        "ANYVAR_AUTH_ISSUER_URL": "https://issuer.example.com",
        "ANYVAR_AUTH_JWKS_URI": "https://example.com/jwks",
    },
)
@patch("anyvar.restapi.auth.is_jwt_token_valid")
async def test_check_bearer_token_falls_back_to_jwt_validation(mock_is_jwt_valid):
    """Test check_bearer_token falls back to JWT when literal token doesn't match."""
    mock_is_jwt_valid.return_value = True

    dependency = get_token_auth_dependency()

    mock_auth = Mock()
    mock_auth.credentials = "not.literal.token"

    result = await dependency(mock_auth)
    assert result is None
    # JWT validation should be called
    mock_is_jwt_valid.assert_called_once()


@pytest.mark.asyncio
@patch.dict(
    os.environ,
    {
        "ANYVAR_AUTH_ISSUER_URL": "https://issuer.example.com",
        "ANYVAR_AUTH_JWKS_URI": "https://example.com/jwks",
    },
)
@patch("anyvar.restapi.auth.is_jwt_token_valid")
async def test_check_bearer_token_exception_handling(mock_is_jwt_valid):
    """Test check_bearer_token converts exceptions to HTTPException."""
    mock_is_jwt_valid.side_effect = Exception("Unexpected error")

    dependency = get_token_auth_dependency()

    mock_auth = Mock()
    mock_auth.credentials = "some.token"

    with pytest.raises(HTTPException) as exc_info:
        await dependency(mock_auth)

    assert exc_info.value.status_code == 401
    assert exc_info.value.detail == "Invalid token"


@pytest.mark.asyncio
@patch.dict(
    os.environ,
    {
        "ANYVAR_AUTH_TOKEN_LIST": "valid_token",
    },
)
async def test_check_bearer_token_with_literal_tokens_only():
    """Test check_bearer_token with only literal tokens configured."""
    dependency = get_token_auth_dependency()

    mock_auth = Mock()
    mock_auth.credentials = "valid_token"

    with patch("anyvar.restapi.auth.is_jwt_token_valid") as mock_is_jwt_valid:
        result = await dependency(mock_auth)
        assert result is None
        # JWT validation should not be called
        mock_is_jwt_valid.assert_not_called()


@pytest.mark.asyncio
@patch.dict(
    os.environ,
    {
        "ANYVAR_AUTH_ISSUER_URL": "https://issuer.example.com",
        "ANYVAR_AUTH_JWKS_URI": "https://example.com/jwks",
    },
)
@patch("anyvar.restapi.auth.is_jwt_token_valid")
async def test_check_bearer_token_with_jwt_only(mock_is_jwt_valid):
    """Test check_bearer_token with only JWT configuration."""
    mock_is_jwt_valid.return_value = True

    dependency = get_token_auth_dependency()

    mock_auth = Mock()
    mock_auth.credentials = "valid.jwt.token"

    result = await dependency(mock_auth)
    assert result is None
    mock_is_jwt_valid.assert_called_once()


@pytest.mark.asyncio
@patch.dict(
    os.environ,
    {
        "ANYVAR_AUTH_TOKEN_LIST": "token1",
        "ANYVAR_AUTH_ISSUER_URL": "https://issuer.example.com",
        "ANYVAR_AUTH_JWKS_URI": "https://example.com/jwks",
    },
)
@patch("anyvar.restapi.auth.is_jwt_token_valid")
async def test_check_bearer_token_both_validations_fail(mock_is_jwt_valid):
    """Test check_bearer_token rejects when both literal and JWT validation fail."""
    mock_is_jwt_valid.return_value = False

    dependency = get_token_auth_dependency()

    mock_auth = Mock()
    mock_auth.credentials = "invalid_token"

    with pytest.raises(HTTPException) as exc_info:
        await dependency(mock_auth)

    assert exc_info.value.status_code == 401
    assert exc_info.value.detail == "Invalid token"



import json
import os
from pathlib import Path

import pytest
from dotenv import load_dotenv
from fastapi.testclient import TestClient
from ga4gh.vrs import models
from pydantic import BaseModel

from anyvar.anyvar import AnyVar, create_storage, create_translator
from anyvar.core import objects
from anyvar.restapi.main import app as anyvar_restapi
from anyvar.restapi.schema import ServiceInfo
from anyvar.storage.base import Storage
from anyvar.translate.base import Translator

pytest_plugins = ("celery.contrib.pytest",)


def pytest_runtest_setup(item):
    """Skip tests not compatible with the current test database backend"""
    all_dbs = {"postgresql", "snowflake"}
    supported_dbs = all_dbs.intersection(mark.name for mark in item.iter_markers())
    current_db = (
        os.environ.get(
            "ANYVAR_TEST_STORAGE_URI",
            "postgresql://postgres:postgres@localhost:5432/anyvar_test",
        )
        .split("+")[0]
        .split(":")[0]
    )
    if supported_dbs and current_db not in supported_dbs:
        pytest.skip(
            f"Skipping test for {supported_dbs} when current test db is {current_db}"
        )


@pytest.fixture(scope="session", autouse=True)
def load_env():
    """Load `.env` file.

    Must set `autouse=True` to run before other fixtures or test cases.
    """
    load_dotenv()


@pytest.fixture(scope="session")
def test_data_dir() -> Path:
    """Provide Path instance pointing to test data directory"""
    return Path(__file__).parent / "data"


@pytest.fixture(scope="session")
def alleles(test_data_dir: Path):
    class _AlleleFixture(BaseModel):
        """Validate data structure in variations.json"""

        variation: models.Allele
        comment: str | None = None
        register_params: dict[str, str | int] | None = None

    with (test_data_dir / "variations.json").open() as f:
        data = json.load(f)
        alleles = data["alleles"]
        for allele in alleles.values():
            assert _AlleleFixture(**allele), f"Not a valid allele fixture: {allele}"
        return alleles


@pytest.fixture(scope="session")
def copy_number_variations(test_data_dir: Path):
    class _CopyNumberFixture(BaseModel):
        """Validate data structure in variations.json"""

        variation: models.CopyNumberChange | models.CopyNumberCount
        comment: str | None = None
        register_params: dict[str, str | int] | None = None

    with (test_data_dir / "variations.json").open() as f:
        data = json.load(f)
        cns = data["copy_numbers"]
        for cn in cns.values():
            assert _CopyNumberFixture(**cn), (
                f"Not a valid copy number variation fixture: {cn}"
            )
        return cns


@pytest.fixture(scope="session")
def celery_config():
    return {
        "broker_url": os.environ.get("CELERY_BROKER_URL", "redis://"),
        "result_backend": os.environ.get("CELERY_BACKEND_URL", "redis://"),
        "task_default_queue": "anyvar_q",
        "event_queue_prefix": "anyvar_ev",
        "task_serializer": "json",
        "result_serializer": "json",
        "accept_content": ["application/json"],
    }


@pytest.fixture(scope="session")
def storage_uri() -> str:
    """Define test storage URI to employ for all storage instance fixtures

    Uses `ANYVAR_TEST_STORAGE_URI` env var
    """
    return os.environ.get(
        "ANYVAR_TEST_STORAGE_URI",
        "postgresql://postgres:postgres@localhost:5432/anyvar_test",
    )


@pytest.fixture(scope="module")
def storage(storage_uri: str):
    """Provide live storage instance from factory.

    Configures from env var ``ANYVAR_TEST_STORAGE_URI``. Defaults to a Postgres DB
    named ``anyvar_test``
    """
    storage = create_storage(uri=storage_uri)
    storage.wipe_db()
    return storage


@pytest.fixture(scope="session")
def translator():
    return create_translator()


@pytest.fixture(scope="module")
def anyvar_instance(storage: Storage, translator: Translator):
    """Provide a test AnyVar instance"""
    return AnyVar(object_store=storage, translator=translator)


@pytest.fixture(scope="module")
def restapi_client(anyvar_instance: AnyVar):
    anyvar_restapi.state.anyvar = anyvar_instance
    anyvar_restapi.state.service_info = ServiceInfo()
    return TestClient(app=anyvar_restapi)


# variation type: VRS-Python model
variation_class_map: dict[str, type[objects.VrsVariation]] = {
    "Allele": models.Allele,
    "CopyNumberCount": models.CopyNumberCount,
    "CopyNumberChange": models.CopyNumberChange,
}


def build_vrs_variant_from_dict(variant_dict: dict) -> objects.VrsVariation:
    """Construct a `VrsVariation` class instance from a dictionary representation of one

    :param variant_dict: a dictionary representation of a `VrsVariation` object
    :return: a `VrsVariation` object
    """
    variant_type = variant_dict.get("type", "")
    return variation_class_map[variant_type](**variant_dict)


@pytest.fixture
def build_vrs_variant_from_dict_function():
    return build_vrs_variant_from_dict



from collections.abc import Callable

import pytest
from ga4gh.vrs import models


@pytest.fixture
def focus_alleles(
    alleles: dict, build_vrs_variant_from_dict_function: Callable[[dict], models.Allele]
) -> tuple[models.Allele, ...]:
    """A small subset of test alleles to use in more focused tests

    This is a tuple because many checks assume a specific order of these objects
    """
    return tuple(
        models.Allele.model_validate(
            build_vrs_variant_from_dict_function(a["variation"])
        )
        for a in (
            alleles["ga4gh:VA.K7akyz9PHB0wg8wBNVlWAAdvMbJUJJfU"],
            alleles["ga4gh:VA.rQBlRht2jfsSp6TpX3xhraxtmgXNKvQf"],
            alleles["ga4gh:VA.1FzYrqG-7jB3Wr46eIL_L5BWElQZEB7i"],
        )
    )


@pytest.fixture
def validated_vrs_alleles(
    alleles: dict, build_vrs_variant_from_dict_function: Callable[[dict], models.Allele]
) -> dict[str, models.Allele]:
    """All allele fixtures, transformed into VRS Pydantic models w/ other test metadata removed"""
    return {
        k: build_vrs_variant_from_dict_function(v["variation"])
        for k, v in alleles.items()
    }



"""Tests postgres storage implementation methods directly."""

import os

import pytest
from ga4gh.vrs import models

from anyvar.storage.postgres import PostgresObjectStore

from .storage_test_funcs import (
    run_alleles_crud,
    run_annotations_crud,
    run_db_lifecycle,
    run_incomplete_objects_error,
    run_mappings_crud,
    run_objects_raises_integrityerror,
    run_query_max_rows,
    run_search_alleles,
    run_sequencelocations_crud,
    run_sequencereferences_crud,
)

pytestmark = pytest.mark.postgresql


@pytest.fixture(scope="session")
def postgres_uri():
    uri = os.environ.get(
        "ANYVAR_TEST_STORAGE_URI",
        "postgresql://postgres:postgres@localhost:5432/anyvar_test",
    )
    return uri


@pytest.fixture
def postgres_storage(postgres_uri: str):
    """Reset storage state after each test case"""
    storage = PostgresObjectStore(postgres_uri)
    yield storage
    storage.wipe_db()


@pytest.mark.ci_ok
def test_db_lifecycle(
    postgres_uri: str, validated_vrs_alleles: dict[str, models.Allele]
):
    # set up and populate DB
    storage = PostgresObjectStore(postgres_uri)
    run_db_lifecycle(storage, validated_vrs_alleles)


@pytest.mark.ci_ok
def test_query_max_rows(
    monkeypatch,
    postgres_storage: PostgresObjectStore,
    focus_alleles: tuple[models.Allele, models.Allele, models.Allele],
):
    run_query_max_rows(monkeypatch, postgres_storage, focus_alleles)


@pytest.mark.ci_ok
def test_alleles_crud(
    postgres_storage: PostgresObjectStore,
    focus_alleles: tuple[models.Allele, models.Allele, models.Allele],
    validated_vrs_alleles: dict[str, models.Allele],
):
    run_alleles_crud(postgres_storage, focus_alleles, validated_vrs_alleles)


@pytest.mark.ci_ok
def test_incomplete_objects_error(postgres_storage: PostgresObjectStore):
    run_incomplete_objects_error(postgres_storage)


@pytest.mark.ci_ok
def test_objects_raises_integrityerror(
    postgres_storage: PostgresObjectStore,
    focus_alleles: tuple[models.Allele, models.Allele, models.Allele],
):
    run_objects_raises_integrityerror(postgres_storage, focus_alleles)


@pytest.mark.ci_ok
def test_sequencelocations_crud(
    postgres_storage: PostgresObjectStore,
    focus_alleles: tuple[models.Allele, models.Allele, models.Allele],
    validated_vrs_alleles: dict[str, models.Allele],
):
    run_sequencelocations_crud(postgres_storage, focus_alleles, validated_vrs_alleles)


@pytest.mark.ci_ok
def test_sequencereferences_crud(
    postgres_storage: PostgresObjectStore,
    focus_alleles: tuple[models.Allele, models.Allele, models.Allele],
    validated_vrs_alleles: dict[str, models.Allele],
):
    run_sequencereferences_crud(postgres_storage, focus_alleles, validated_vrs_alleles)


@pytest.mark.ci_ok
def test_mappings_crud(
    postgres_storage: PostgresObjectStore,
    validated_vrs_alleles: dict[str, models.Allele],
):
    run_mappings_crud(postgres_storage, validated_vrs_alleles)


@pytest.mark.ci_ok
def test_annotations_crud(
    postgres_storage: PostgresObjectStore,
    focus_alleles: tuple[models.Allele, models.Allele, models.Allele],
):
    run_annotations_crud(postgres_storage, focus_alleles)


def test_search_alleles(
    postgres_storage: PostgresObjectStore,
    validated_vrs_alleles: dict[str, models.Allele],
):
    run_search_alleles(postgres_storage, validated_vrs_alleles)



"""Normalize incoming variation descriptions with the VRS-Python library."""

from os import environ

from ga4gh.vrs import models
from ga4gh.vrs.dataproxy import _DataProxy, create_dataproxy
from ga4gh.vrs.extras.translator import AlleleTranslator, CnvTranslator

from anyvar.core.objects import VrsVariation
from anyvar.restapi.schema import SupportedVariationType
from anyvar.translate.base import TranslationError, Translator


class VrsPythonTranslator(Translator):
    """Translator layer using VRS-Python Translator class."""

    def __init__(self, seqrepo_proxy: _DataProxy | None = None) -> None:
        """Initialize VRS-Python translator.

        If an existing SeqRepo data proxy is not provided, use the VRS-Python
        ``create_dataproxy`` function to construct one. The following URI patterns
        can be used to specify a local file instance or an address to REST service:

        * seqrepo+file:///path/to/seqrepo/root
        * seqrepo+:../relative/path/to/seqrepo/root
        * seqrepo+http://localhost:5000/seqrepo
        * seqrepo+https://somewhere:5000/seqrepo

        We'll pass the value defined under the environment variable
        ``SEQREPO_DATAPROXY_URI``, and default to local service
        (``"seqrepo+http://localhost:5000/seqrepo"``) if it's undefined.

        :param seqrepo_proxy: existing SR proxy instance if available.
        """
        if not seqrepo_proxy:
            seqrepo_uri = environ.get(
                "SEQREPO_DATAPROXY_URI", "seqrepo+http://localhost:5000/seqrepo"
            )
            self.dp = create_dataproxy(seqrepo_uri)
        self.allele_tlr = AlleleTranslator(data_proxy=self.dp)
        self.cnv_tlr = CnvTranslator(data_proxy=self.dp)

    def translate_variation(self, var: str, **kwargs) -> VrsVariation:
        """Translate provided variation text into a VRS Variation object.

        :param var: user-provided string describing or referencing a variation.
        :param input_type: The type of variation for `var`.
        :keyword types.VrsVariation input_type: The type of variation for `var`. If
            not provided, will first try to translate to allele and then copy number
        :keyword int copies: The number of copies for VRS Copy Number Count
        :keyword models.CopyChange copy_change: The EFO code for VRS COpy Number Change
        :keyword ReferenceAssembly assembly_name: Assembly name for ``var``.
            Only used when ``var`` uses gnomad format.
        :returns: VRS variation object
        :raises TranslationError: if translation is unsuccessful, either because
            the submitted variation is malformed, or because VRS-Python doesn't support
            its translation.
        """
        input_type = kwargs.get("input_type")
        if input_type == SupportedVariationType.ALLELE:
            variation = self.translate_allele(var, **kwargs)
        elif input_type in (
            SupportedVariationType.COPY_NUMBER_CHANGE,
            SupportedVariationType.COPY_NUMBER_COUNT,
        ):
            variation = self.translate_cnv(var, **kwargs)
        else:
            # Try allele then copy number
            try:
                variation = self.translate_allele(var)
            except TranslationError:
                try:
                    variation = self.translate_cnv(var, **kwargs)
                except TranslationError as e:
                    msg = f"{var} isn't supported by the VRS-Python AlleleTranslator or CnvTranslator."
                    raise TranslationError(msg) from e

        return variation

    def translate_allele(self, var: str, **kwargs) -> models.Allele:
        """Translate provided variation text into a VRS Allele object.

        :param var: user-provided string describing or referencing a variation.
        :kwargs:
            assembly_name(str) -> Assembly name for ``var``.
            Only used when ``var`` uses gnomad format.
            Defaults to "GRCh38". Must be "GRCh38" or "GRCh7"
            VRS-Python sets a default, but we should set a default just in case
            VRS-Python ever changes the default.
        :returns: VRS variation object if able to translate
        :raises TranslationError: if translation is unsuccessful, either because
            the submitted variation is malformed, or because VRS-Python doesn't support
            its translation.
        """
        try:
            return self.allele_tlr.translate_from(var, fmt=None, **kwargs)  # type: ignore (this will always return a models.Allele instance, or raise a ValueError)
        except ValueError as e:
            msg = f"{var} isn't supported by the VRS-Python AlleleTranslator."
            raise TranslationError(msg) from e

    def translate_cnv(
        self, var: str, **kwargs
    ) -> models.CopyNumberCount | models.CopyNumberChange:
        """Translate provided variation text into a VRS object.

        :param var: user-provided string describing or referencing a variation.
        :kwargs:
            copies(int) - The number of copies for VRS Copy Number Count
            copy_change (models.CopyChange) - The EFO code for VRS COpy Number Change
        :returns: VRS variation object if able to translate
        :raises TranslationError: if translation is unsuccessful, either because
            the submitted variation is malformed, or because VRS-Python doesn't support
            its translation.
        """
        try:
            return self.cnv_tlr.translate_from(var, fmt=None, **kwargs)  # type: ignore (this will always return a models.CopyNumberCount | models.CopyNumberChange instance, or raise a ValueError)
        except ValueError as e:
            msg = f"{var} isn't supported by the VRS-Python CnvTranslator."
            raise TranslationError(msg) from e

    def get_sequence_id(self, accession_id: str) -> str:
        """Get GA4GH sequence identifier for provided accession ID

        :param accession_id: ID to convert
        :return: equivalent GA4GH sequence ID
        :raise: KeyError if no equivalent ID is available
        """
        result = self.allele_tlr.data_proxy.translate_sequence_identifier(
            accession_id, "ga4gh"
        )
        return result[0]



from typing import get_args

import pytest

from anyvar.core.objects import VrsVariation
from anyvar.restapi.schema import SupportedVariationType
from anyvar.storage.orm import _camel_to_snake


@pytest.mark.ci_ok
def test_supported_object_type():
    for cls in get_args(VrsVariation):
        assert SupportedVariationType[_camel_to_snake(cls.__name__)] == cls.__name__
    assert len(get_args(VrsVariation)) == len(SupportedVariationType)



# /// script  # noqa: D100
# dependencies = [
#   "requests",
#   "click",
# ]
#
# invoke with e.g. `uv run scripts/vcf_benchmark.py --workers 2 vcfs/vcf_1.vcf vcfs/vcf_2.vcf vcfs/vcf_3.vcf vcfs/vcf_4.vcf`
# ///

import logging
import time
from concurrent.futures import ProcessPoolExecutor
from enum import StrEnum
from functools import partial
from json import JSONDecodeError
from pathlib import Path
from timeit import default_timer as timer

import click
import requests

LOG_FILE = "anyvar_benchmark.log"
logging.basicConfig(
    level=logging.INFO,
    format="%(asctime)s [%(levelname)s] %(message)s",
    handlers=[
        logging.FileHandler(LOG_FILE),
        logging.StreamHandler(),  # still output to console
    ],
)
logger = logging.getLogger(__name__)

HTTP_TIMEOUT = 120


def submit_variants(
    file: Path,
    anyvar_host: str,
    for_ref: bool,
    allow_async_write: bool,
    assembly: str,
    run_async: bool,
) -> None:
    with file.open("rb") as f:
        response = requests.put(
            f"{anyvar_host}/vcf",
            files={"vcf": (file.name, f, "text/plain")},
            params={
                "for_ref": for_ref,
                "allow_async_write": allow_async_write,
                "assembly": assembly,
                "run_async": run_async,
            },
            timeout=HTTP_TIMEOUT,
            headers={"accept": "application/json"},
        )
    logger.info("Submitting VCF to ingestion endpoint")
    response.raise_for_status()
    if run_async:
        run_id = response.json()["run_id"]
        logger.info("Submission successful for %s", run_id)
        while True:
            logger.info("Polling ingestion status...")
            time.sleep(5)
            response = requests.get(f"{anyvar_host}/vcf/{run_id}", timeout=HTTP_TIMEOUT)
            try:
                if response.json()["status"] != "PENDING":
                    logger.info("Ingestion complete!")
                    break
            except JSONDecodeError:
                # presumably a successful response
                if response.text.startswith("##fileformat="):
                    logger.info("Annotated VCF returned successfully")
                else:
                    filename = f"vcf_benchmark_output_{run_id}"
                    with Path(filename).open("w") as f:
                        f.write(response.text)
                    logger.info(
                        "Received malformed response from vcf run ID lookup. Saved to %s",
                        filename,
                    )
                return
    # don't need to poll for sync response
    # timeouts should be expected for larger files though
    if response.text.startswith("##fileformat="):
        logger.info("Annotated VCF returned successfully")
    else:
        logger.info("Something has gone wrong: %s", response.text)


class AssemblyOption(StrEnum):
    GRCH38 = "GRCh38"
    GRCH37 = "GRCh37"


def process_file(
    filepath: Path,
    *,
    for_ref: bool,
    allow_async_write: bool,
    assembly: str,
    run_async: bool,
) -> None:
    anyvar_host = "http://localhost:8000"
    start = timer()
    submit_variants(
        filepath, anyvar_host, for_ref, allow_async_write, assembly, run_async
    )
    end = timer()
    logger.info("Completed VCF submission of %s in %s", filepath, f"{end - start:.5f}")


@click.command()
@click.argument("filepaths", type=click.Path(exists=True, path_type=Path), nargs=-1)
@click.option(
    "--workers",
    type=int,
    default=4,
    show_default=True,
    help="Number of parallel workers",
)
@click.option("--for-ref", type=bool, default=True)
@click.option("--allow-async-write", type=bool, default=True)
@click.option("--assembly", type=click.Choice(AssemblyOption, case_sensitive=False))
@click.option("--run-async", type=bool, default=True)
def main(
    filepaths: tuple[Path],
    workers: int,
    for_ref: bool,
    allow_async_write: bool,
    assembly: str,
    run_async: bool,
):
    logger.info("Received %s file(s):", len(filepaths))
    for fp in filepaths:
        logger.info("- %s", fp)
    logger.info("Using %s workers", workers)
    start = timer()
    _process_file = partial(
        process_file,
        for_ref=for_ref,
        allow_async_write=allow_async_write,
        assembly=assembly,
        run_async=run_async,
    )
    with ProcessPoolExecutor(max_workers=workers) as executor:
        results = list(executor.map(_process_file, filepaths))
    for result in results:
        logger.info(result)

    end = timer()
    duration = end - start
    logger.info("Completed ingest of %s in %s", filepaths, f"{duration:.5f}")


if __name__ == "__main__":
    main()



"""Define types and classes for *metadata*, i.e. additional descriptions on top of stored objects.

This currently includes

* **Annotations**: free-text key-value descriptions of an object
* **Mappings**: directed associations between objects
"""

from enum import StrEnum

from pydantic import BaseModel, JsonValue


class VariationMappingType(StrEnum):
    """Supported mapping types between VRS Variations."""

    LIFTOVER = "liftover"
    TRANSCRIPTION = "transcription"
    TRANSLATION = "translation"


class VariationMapping(BaseModel):
    """Describe a mapping between two variations."""

    source_id: str
    dest_id: str
    mapping_type: VariationMappingType


class AnnotationType(StrEnum):
    """Describe commonly used annotation types"""

    CREATION_TIMESTAMP = "creation_timestamp"


class Annotation(BaseModel):
    """Generic annotation class which attaches any object to an identifier"""

    object_id: str
    annotation_type: str
    annotation_value: JsonValue



import contextlib
import os
import shutil

import pytest
from celery.contrib.testing.worker import start_worker
from celery.result import AsyncResult
from pytest_mock import MockerFixture
from tests.conftest import build_vrs_variant_from_dict

import anyvar.anyvar
from anyvar.queueing.celery_worker import celery_app
from anyvar.storage.base import Storage


@pytest.fixture
def preloaded_alleles(storage: Storage, alleles: dict):
    """Provide alleles that have been loaded into the storage instance for this scope.

    Utilizing this fixture means that everything in the alleles fixture will be preloaded
    into the storage instance that the test function ultimately receives.
    """
    storage.wipe_db()
    storage.add_objects(
        [build_vrs_variant_from_dict(a["variation"]) for a in alleles.values()]
    )
    return alleles


@pytest.fixture(scope="session")
def vcf_run_id() -> str:
    """Static run ID to use in VCF tests"""
    return "1234"


@pytest.fixture
def celery_context(mocker: MockerFixture, vcf_run_id: str, storage_uri: str):
    """Provide setup/teardown for test cases that use celery workers for VCF operations

    Because the Celery worker doesn't currently take an injectable storage class instance,
    we have to mock the `ANYVAR_STORAGE_URI` connection string to use the test database
    """
    assert anyvar.anyvar.has_queueing_enabled(), "async VCF queueing is not enabled"
    mocker.patch.dict(
        os.environ,
        {
            "ANYVAR_STORAGE_URI": storage_uri,
            "ANYVAR_VCF_ASYNC_WORK_DIR": "tests/tmp_async_work_dir",
        },
    )
    context_manager = start_worker(
        celery_app,
        pool="solo",
        loglevel="info",
        perform_ping_check=False,
        shutdown_timeout=30,
    )
    context_manager.__enter__()
    celery_app.control.purge()
    AsyncResult(vcf_run_id).forget()

    yield

    context_manager.__exit__(None, None, None)
    with contextlib.suppress(FileNotFoundError):
        # a test probably failed early
        shutil.rmtree("tests/tmp_async_work_dir")



"""Provide base properties for Translator classes."""

from abc import ABC, abstractmethod

from ga4gh.vrs.dataproxy import _DataProxy

from anyvar.core.objects import VrsVariation


class TranslatorSetupError(Exception):
    """Indicates failure to create translator instance (e.g. invalid params provided)"""


class TranslatorConnectionError(Exception):
    """Indicates failure to connect to translator instance (e.g. REST endpoint not
    responding)
    """


class TranslationError(Exception):
    """Indicates failure to translate provided term into known variation structure."""


class Translator(ABC):
    """Base Translator class.

    Use for
    * Translating incoming variant expressions to VRS
    * Acquiring reference accessions for incoming sequence descriptions/references

    """

    dp: _DataProxy

    @abstractmethod
    def translate_variation(self, var: str, **kwargs) -> VrsVariation:
        """Translate provided variation text into a VRS Variation object.

        :param var: user-provided string describing or referencing a variation.
        :kwargs:
           * input_type (types.VrsVariation): The type of variation for `var`. If
             not provided, will first try to translate to allele and then copy number
           * copies (int) - The number of copies for VRS Copy Number Count
           * copy_change (models.CopyChange) - The EFO code for VRS COpy Number Change
           * assembly_name(str) -> Assembly name for ``var``. Only used when ``var``
             uses gnomad format. Defaults to "GRCh38". Must be "GRCh38" or "GRCh37".
             VRS-Python sets a default, but we should set a default just in case VRS-Python
             ever changes the default.
        :returns: VRS variation object
        :raises TranslationError: if translation is unsuccessful, either because
            the submitted variation is malformed, or because VRS-Python doesn't support
            its translation.
        """
        raise NotImplementedError

    @abstractmethod
    def translate_allele(self, var: str) -> VrsVariation | None:
        """Translate provided variation text into a normalized VRS object.

        :param var: user-provided string describing or referencing a variation.
        :returns: VRS variation object if able to normalize
        :raises TranslatorConnectionError: if translation request returns error
        """
        raise NotImplementedError

    @abstractmethod
    def translate_cnv(self, var: str) -> VrsVariation | None:
        """Translate provided variation text into a normalized VRS object.

        :param var: user-provided string describing or referencing a variation.
        :returns: VRS variation object if able to normalize
        :raises TranslatorConnectionError: if translation request returns error
        """
        raise NotImplementedError

    @abstractmethod
    def get_sequence_id(self, accession_id: str) -> str:
        """Get GA4GH sequence identifier for provided accession ID

        :param accession_id: ID to convert
        :return: equivalent GA4GH sequence ID
        :raise: KeyError if no equivalent ID is available
        """
        raise NotImplementedError



"""Central registry for all object mappers."""

from typing import TypeVar

from ga4gh.vrs import models as vrs_models

from anyvar.core import metadata
from anyvar.storage import orm
from anyvar.storage.mappers import (
    AlleleMapper,
    AnnotationMapper,
    BaseMapper,
    SequenceLocationMapper,
    SequenceReferenceMapper,
    VariationMappingMapper,
)

T = TypeVar("T")


class MapperRegistry:
    """Central registry for all object mappers."""

    def __init__(self) -> None:
        """Initialize the MapperRegistry with known mappers."""
        self.anyvar_to_db_mapping = {
            vrs_models.Allele: orm.Allele,
            vrs_models.SequenceLocation: orm.Location,
            vrs_models.SequenceReference: orm.SequenceReference,
            metadata.VariationMapping: orm.VariationMapping,
            metadata.Annotation: orm.Annotation,
        }

        self._mappers: dict[type, BaseMapper] = {
            orm.Allele: AlleleMapper(),
            orm.Location: SequenceLocationMapper(),
            orm.SequenceReference: SequenceReferenceMapper(),
            orm.VariationMapping: VariationMappingMapper(),
            orm.Annotation: AnnotationMapper(),
        }

    def get_mapper(self, entity_type: type[T]) -> BaseMapper:
        """Get mapper for the given entity type."""
        mapper = self._mappers.get(entity_type)
        if mapper is None:
            raise ValueError(f"No mapper registered for type: {entity_type}")
        return mapper

    def from_db_entity(self, db_entity):  # noqa: ANN201, ANN001
        """Convert any DB entity to its corresponding VRS model."""
        mapper = self.get_mapper(type(db_entity))
        return mapper.from_db_entity(db_entity)

    def to_db_entity(self, anyvar_entity) -> orm.Base:  # noqa: ANN001
        """Convert any VRS model to its corresponding DB entity."""
        # Map VRS model types to DB entity types

        db_type = self.anyvar_to_db_mapping.get(type(anyvar_entity))
        if db_type is None:
            raise ValueError(
                f"No DB entity type mapped for VRS model: {type(anyvar_entity)}"
            )

        mapper = self.get_mapper(db_type)
        return mapper.to_db_entity(anyvar_entity)


# Global registry instance
mapper_registry = MapperRegistry()



"""Provide PostgreSQL-based storage implementation."""

import json
import logging
from collections import defaultdict
from collections.abc import Iterable

from ga4gh.vrs import models as vrs_models
from sqlalchemy import create_engine, delete, func, select
from sqlalchemy.dialects.postgresql import insert
from sqlalchemy.exc import IntegrityError
from sqlalchemy.orm import joinedload, sessionmaker

from anyvar.core import metadata
from anyvar.core import objects as anyvar_objects
from anyvar.storage import orm
from anyvar.storage.base import (
    DataIntegrityError,
    IncompleteVrsObjectError,
    InvalidSearchParamsError,
    Storage,
)
from anyvar.storage.mapper_registry import mapper_registry
from anyvar.storage.orm import create_tables

_logger = logging.getLogger(__name__)


class PostgresObjectStore(Storage):
    """PostgreSQL storage backend using dedicated ORM tables.

    This implementation uses the new Allele, Location, and SequenceReference tables
    with object mapping to convert between VRS models and database entities.
    """

    # temporary cap on max # of rows that can be returned by a single SQL query
    # issue 295 should convert this to a batch size parameter
    MAX_ROWS = 100

    _VRS_OBJECT_INSERT_ORDER: list[str] = [  # noqa: RUF012
        orm.SequenceReference.__name__,
        orm.Location.__name__,
        orm.Allele.__name__,
    ]

    def __init__(self, db_url: str, *args, **kwargs) -> None:
        """Initialize PostgreSQL storage.

        :param db_url: Database connection URL (e.g., postgresql://user:pass@host:port/db)
        """
        self.db_url = db_url
        self.engine = create_engine(db_url)
        self.session_factory = sessionmaker(bind=self.engine)
        create_tables(self.db_url)

    def close(self) -> None:
        """Close the storage backend."""

    def wait_for_writes(self) -> None:
        """Wait for all background writes to complete.
        NOTE: This is a no-op for synchronous storage backends.
        """

    def wipe_db(self) -> None:
        """Wipe all data from the storage backend."""
        with self.session_factory() as session, session.begin():
            # Delete all data from tables in dependency order
            session.execute(delete(orm.VariationMapping))
            session.execute(delete(orm.Allele))
            session.execute(delete(orm.Location))
            session.execute(delete(orm.SequenceReference))

            # Delete other tables
            session.execute(delete(orm.VrsObject))
            session.execute(delete(orm.Annotation))

    def add_objects(self, objects: Iterable[anyvar_objects.VrsObject]) -> None:
        """Add multiple VRS objects to storage using bulk inserts.

        If an object ID conflicts with an existing object, skip it.

        This method assumes that for VRS objects (e.g. `Allele`, `SequenceLocation`,
        `SequenceReference`) the `.id` property is present and uses the correct
        GA4GH identifier for that object. It also assumes that contained objects are
        similarly properly identified and materialized in full, not just as an IRI reference.
        An error is raised if these assumptions are violated, rolling back the entire
        transaction.

        :param objects: VRS objects to add to storage
        :raise IncompleteVrsObjectError: if object is missing required properties or if
            required properties aren't fully dereferenced
        """
        objects_list: list[anyvar_objects.VrsObject] = list(objects)
        if not objects_list:
            return

        # Collect unique entities by ID to avoid duplicates
        vrs_objects = defaultdict(dict[str, orm.Base])

        # Process all objects and extract their components
        for vrs_object in objects_list:
            try:
                db_entity = mapper_registry.to_db_entity(vrs_object)
            except AttributeError as e:
                raise IncompleteVrsObjectError from e

            object_parts = db_entity.disassemble()
            for entity_type, entity in object_parts.items():
                vrs_objects[entity_type][entity.id] = entity  # type: ignore (all children of orm.Base have an `id`)

        with self.session_factory() as session, session.begin():
            # Use ON CONFLICT DO NOTHING to handle duplicates gracefully
            # We should have already de-duplicated by ID above, but duplicates
            # may already exist in the database.

            for vrs_object_type in self._VRS_OBJECT_INSERT_ORDER:
                objects_by_id = vrs_objects[vrs_object_type]
                if objects_by_id:
                    dicts = [entity.to_dict() for entity in objects_by_id.values()]
                    orm_model = getattr(orm, vrs_object_type)
                    stmt = insert(orm_model).on_conflict_do_nothing()
                    session.execute(stmt, dicts)

    def get_objects(
        self, object_type: type[anyvar_objects.VrsObject], object_ids: Iterable[str]
    ) -> Iterable[anyvar_objects.VrsObject]:
        """Retrieve multiple VRS objects from storage by their IDs.

        If no object matches a given ID, that ID is skipped

        :param object_type: type of object to get
        :param object_ids: IDs of objects to fetch
        :return: iterable collection of VRS objects matching given IDs
        """
        object_ids_list = list(object_ids)
        results = []

        with self.session_factory() as session:
            if object_type is vrs_models.Allele:
                # Get alleles with eager loading
                stmt = (
                    select(orm.Allele)
                    .options(
                        joinedload(orm.Allele.location).joinedload(
                            orm.Location.sequence_reference
                        )
                    )
                    .where(orm.Allele.id.in_(object_ids_list))
                    .limit(self.MAX_ROWS)
                )
                db_objects = session.scalars(stmt).all()
            elif object_type is vrs_models.SequenceLocation:
                # Get locations with eager loading
                stmt = (
                    select(orm.Location)
                    .options(joinedload(orm.Location.sequence_reference))
                    .where(orm.Location.id.in_(object_ids_list))
                    .limit(self.MAX_ROWS)
                )
                db_objects = session.scalars(stmt).all()
            elif object_type is vrs_models.SequenceReference:
                # Get sequence references
                stmt = (
                    select(orm.SequenceReference)
                    .where(orm.SequenceReference.id.in_(object_ids_list))
                    .limit(self.MAX_ROWS)
                )
                db_objects = session.scalars(stmt).all()
            else:
                raise ValueError(f"Unsupported object type: {object_type}")

            for db_object in db_objects:
                vrs_object = mapper_registry.from_db_entity(db_object)
                results.append(vrs_object)

        return results

    def delete_objects(
        self, object_type: type[anyvar_objects.VrsObject], object_ids: Iterable[str]
    ) -> None:
        """Delete all objects of a specific type from storage.

        * If no object matching a given ID is found, it's ignored.
        * Deletes do not cascade.

        :param object_type: type of objects to delete
        :param object_ids: IDs of objects to delete
        :raise DataIntegrityError: if attempting to delete an object which is
            depended upon by another object
        """
        object_ids_list = list(object_ids)

        with self.session_factory() as session, session.begin():
            if object_type is vrs_models.Allele:
                stmt = delete(orm.Allele).where(orm.Allele.id.in_(object_ids_list))
            elif object_type is vrs_models.SequenceLocation:
                stmt = delete(orm.Location).where(orm.Location.id.in_(object_ids_list))
            elif object_type is vrs_models.SequenceReference:
                stmt = delete(orm.SequenceReference).where(
                    orm.SequenceReference.id.in_(object_ids_list)
                )
            else:
                raise ValueError(f"Unsupported object type: {object_type}")
            try:
                session.execute(stmt)
            except IntegrityError as e:
                _logger.exception(
                    "Attempted deletion that violated a foreign key constraint"
                )
                raise DataIntegrityError from e

    def add_mapping(self, mapping: metadata.VariationMapping) -> None:
        """Add a mapping between two objects.

        If the mapping instance already exists, do nothing.

        Todo:
        * Implement insert constraint/MissingVariationReferenceError in #286

        :param mapping: mapping object
        :raises ValueError: If source_id equals dest_id
        :raise MissingVariationReferenceError: if source or destination IDs aren't present in DB

        """
        if mapping.source_id == mapping.dest_id:
            msg = f"source_id cannot equal dest_id: {mapping.source_id}"
            raise ValueError(msg)

        stmt = (
            insert(orm.VariationMapping)
            .values(
                [
                    {
                        "source_id": mapping.source_id,
                        "dest_id": mapping.dest_id,
                        "mapping_type": mapping.mapping_type,
                    }
                ]
            )
            .on_conflict_do_nothing()
        )
        try:
            with self.session_factory() as session, session.begin():
                session.execute(stmt)
        except IntegrityError as e:
            raise KeyError from e

    def delete_mapping(self, mapping: metadata.VariationMapping) -> None:
        """Delete a mapping between two objects.

        * If no such mapping exists in the DB, does nothing.
        * Deletes do not cascade.

        :param mapping: mapping object
        """
        stmt = (
            delete(orm.VariationMapping)
            .where(orm.VariationMapping.source_id == mapping.source_id)
            .where(orm.VariationMapping.dest_id == mapping.dest_id)
            .where(orm.VariationMapping.mapping_type == mapping.mapping_type)
        )
        with self.session_factory() as session, session.begin():
            session.execute(stmt)

    def get_mappings(
        self,
        source_object_id: str,
        mapping_type: metadata.VariationMappingType | None = None,
    ) -> Iterable[metadata.VariationMapping]:
        """Return an iterable of mappings from the source ID

        Optionally provide a type to filter results.

        :param source_object_id: ID of the source object
        :param mapping_type: The type of mapping to retrieve (defaults to `None` to
            retrieve all mappings for the source ID)
        :return: iterable collection of mapping descriptors (empty if no matching mappings exist)
        """
        stmt = (
            select(orm.VariationMapping)
            .where(orm.VariationMapping.source_id == source_object_id)
            .limit(self.MAX_ROWS)
        )
        if mapping_type:
            stmt = stmt.where(orm.VariationMapping.mapping_type == mapping_type)
        with self.session_factory() as session, session.begin():
            mappings = session.scalars(stmt).all()
            return [mapper_registry.from_db_entity(mapping) for mapping in mappings]

    def add_annotation(self, annotation: metadata.Annotation) -> None:
        """Add an annotation to the database.

        Adding the same annotation repeatedly creates redundant records.

        Todo:
        * Implement insert constraint/MissingVariationReferenceError in #286

        :param annotation: The annotation to add
        :raise MissingVariationReferenceError: if no object corresponding to the annotation's object ID is present in DB

        """
        db_entity: orm.Annotation = mapper_registry.to_db_entity(annotation)
        stmt = insert(orm.Annotation).returning(orm.Annotation.id)
        with self.session_factory() as session, session.begin():
            session.execute(stmt, db_entity.to_dict()).scalar_one()

    def get_annotations(
        self, object_id: str, annotation_type: str | None = None
    ) -> list[metadata.Annotation]:
        """Get all annotations for the specified object, optionally filtered by type.

        :param object_id: The ID of the object to retrieve annotations for
        :param annotation_type: The type of annotation to retrieve (defaults to `None` to retrieve all annotations for the object)
        :return: A list of annotations
        """
        stmt = select(orm.Annotation).where(orm.Annotation.object_id == object_id)

        if annotation_type:
            stmt = stmt.where(orm.Annotation.annotation_type == annotation_type)

        stmt = stmt.limit(self.MAX_ROWS)

        with self.session_factory() as session, session.begin():
            db_annotations = session.execute(stmt).scalars().all()

            return [
                mapper_registry.from_db_entity(db_annotation)
                for db_annotation in db_annotations
            ]

    def delete_annotation(self, annotation: metadata.Annotation) -> None:
        """Deletes an annotation from the database

        * If no such annotation exists, do nothing.
        * Deletes do not cascade.

        :param annotation: The annotation object to delete
        """
        stmt = (
            delete(orm.Annotation)
            .where(orm.Annotation.object_id == annotation.object_id)
            .where(orm.Annotation.annotation_type == annotation.annotation_type)
            .where(
                orm.Annotation.annotation_value
                == json.dumps(annotation.annotation_value)
            )
        )
        with self.session_factory() as session, session.begin():
            session.execute(stmt)

    def search_alleles(
        self,
        refget_accession: str,
        start: int,
        stop: int,
    ) -> list[vrs_models.Allele]:
        """Find all Alleles that are located within the specified interval.

        The interval is the closed range [start, stop] on the sequence identified by
        the RefGet SequenceReference accession (`SQ.*`). Both `start` and `stop` are
        inclusive and represent inter-residue positions.

        Currently, any variation which overlaps the queried region is returned.

        Todo (see Issue #338):
        * define alternate match modes (partial/full overlap/contained/etc)
        * define behavior for LSE indels and for alternative types of state (RLEs)

        Raises an error if
        * `start` or `end` are negative
        * `end` > `start`

        :param refget_accession: refget accession (e.g. `"SQ.IW78mgV5Cqf6M24hy52hPjyyo5tCCd86"`)
        :param start: Inclusive, inter-residue start position of the interval
        :param stop: Inclusive, inter-residue end position of the interval
        :return: a list of matching VRS alleles
        :raise InvalidSearchParamsError: if above search param requirements are violated

        """
        if start < 0 or stop < 0 or start > stop:
            raise InvalidSearchParamsError

        with self.session_factory() as session:
            # Query alleles with overlapping locations
            # NOTE: this is any overlap, not containment.
            stmt = (
                select(orm.Allele)
                .options(
                    joinedload(orm.Allele.location).joinedload(
                        orm.Location.sequence_reference
                    )
                )
                .join(orm.Location)
                .join(orm.SequenceReference)
                .where(
                    orm.SequenceReference.id == refget_accession,
                    func.int8range(orm.Location.start, orm.Location.end, "[]").op("&&")(
                        func.int8range(start, stop, "[]")
                    ),
                )
                .limit(self.MAX_ROWS)
            )
            db_alleles = session.scalars(stmt).all()

            return [
                mapper_registry.from_db_entity(db_allele) for db_allele in db_alleles
            ]



"""Test VCF input/output features."""

import io
import os
import pathlib
import time
from http import HTTPStatus

import pytest
from billiard.exceptions import TimeLimitExceeded
from celery.exceptions import WorkerLostError
from fastapi.testclient import TestClient
from pytest_mock import MockerFixture

from anyvar.restapi.vcf_router import _working_file_cleanup


@pytest.fixture
def sample_vcf_grch38():
    """Basic GRCh38 VCF fixture to use in tests."""
    file_content = b"""##fileformat=VCFv4.2
##hailversion=0.2.100-2ea2615a797a
##INFO=<ID=QUALapprox,Number=1,Type=Integer,Description="">
##INFO=<ID=SB,Number=.,Type=Integer,Description="">
##INFO=<ID=MQ,Number=1,Type=Float,Description="">
##INFO=<ID=MQRankSum,Number=1,Type=Float,Description="">
##INFO=<ID=VarDP,Number=1,Type=Integer,Description="">
##INFO=<ID=AS_ReadPosRankSum,Number=1,Type=Float,Description="">
##INFO=<ID=AS_pab_max,Number=1,Type=Float,Description="">
##INFO=<ID=AS_QD,Number=1,Type=Float,Description="">
##INFO=<ID=AS_MQ,Number=1,Type=Float,Description="">
##INFO=<ID=QD,Number=1,Type=Float,Description="">
##INFO=<ID=AS_MQRankSum,Number=1,Type=Float,Description="">
##INFO=<ID=FS,Number=1,Type=Float,Description="">
##INFO=<ID=AS_FS,Number=1,Type=Float,Description="">
##INFO=<ID=ReadPosRankSum,Number=1,Type=Float,Description="">
##INFO=<ID=AS_QUALapprox,Number=1,Type=Integer,Description="">
##INFO=<ID=AS_SB_TABLE,Number=.,Type=Integer,Description="">
##INFO=<ID=AS_VarDP,Number=1,Type=Integer,Description="">
##INFO=<ID=AS_SOR,Number=1,Type=Float,Description="">
##INFO=<ID=SOR,Number=1,Type=Float,Description="">
##INFO=<ID=singleton,Number=0,Type=Flag,Description="">
##INFO=<ID=transmitted_singleton,Number=0,Type=Flag,Description="">
##INFO=<ID=omni,Number=0,Type=Flag,Description="">
##INFO=<ID=mills,Number=0,Type=Flag,Description="">
##INFO=<ID=monoallelic,Number=0,Type=Flag,Description="">
##INFO=<ID=AS_VQSLOD,Number=1,Type=Float,Description="">
##INFO=<ID=InbreedingCoeff,Number=1,Type=Float,Description="">
##FILTER=<ID=AC0,Description="Allele count is zero after filtering out low-confidence genotypes (GQ < 20; DP < 10; and AB < 0.2 for het calls)">
##FILTER=<ID=AS_VQSR,Description="Failed VQSR filtering thresholds of -2.7739 for SNPs and -1.0606 for indels">
##contig=<ID=chr1,length=248956422,assembly=GRCh38>
#CHROM	POS	ID	REF	ALT	QUAL	FILTER	INFO
chr1	10330	.	CCCCTAACCCTAACCCTAACCCTACCCTAACCCTAACCCTAACCCTAACCCTAA	C	.	PASS	QUALapprox=21493;SB=325,1077,113,694;MQ=32.1327;MQRankSum=0.720000;VarDP=2236;AS_ReadPosRankSum=-0.736000;AS_pab_max=1.00000;AS_QD=5.17857;AS_MQ=29.5449;QD=9.61225;AS_MQRankSum=0.00000;FS=8.55065;AS_FS=.;ReadPosRankSum=0.727000;AS_QUALapprox=145;AS_SB_TABLE=325,1077,2,5;AS_VarDP=28;AS_SOR=0.311749;SOR=1.48100;singleton;AS_VQSLOD=13.4641;InbreedingCoeff=-0.000517845"""
    return io.BytesIO(file_content)


def test_vcf_registration_default_assembly(
    restapi_client: TestClient, sample_vcf_grch38: io.BytesIO
):
    """Test registration and annotation of VCFs with default assembly"""
    resp = restapi_client.put("/vcf", files={"vcf": ("test.vcf", sample_vcf_grch38)})
    assert resp.status_code == HTTPStatus.OK
    assert (
        b"VRS_Allele_IDs=ga4gh:VA.5PqxTNMJZYJqQZ8MgF_77I1I_qcddGN_,ga4gh:VA._QhHH18HBAIeLos6npRgR-S_0lAX5KR6"
        in resp.content
    )


def test_vcf_registration_vrs_attrs(
    restapi_client: TestClient, sample_vcf_grch38: io.BytesIO
):
    """Test optional inclusion of VRS attrs"""
    resp = restapi_client.put(
        "/vcf",
        files={"vcf": ("test.vcf", sample_vcf_grch38)},
        params={"add_vrs_attributes": True},
    )
    assert resp.status_code == HTTPStatus.OK
    assert (
        b"VRS_Starts=10329,10330;VRS_Ends=10383,10392;VRS_States=,CCCTAACCC;VRS_Lengths=54,9;VRS_RepeatSubunitLengths=54,53"
        in resp.content
    )


def test_vcf_registration_grch38(
    restapi_client: TestClient, sample_vcf_grch38: io.BytesIO
):
    """Test registration and annotation of VCFs specifying GRCh38 as the assembly to use"""
    resp = restapi_client.put(
        "/vcf",
        params={"assembly": "GRCh38"},
        files={"vcf": ("test.vcf", sample_vcf_grch38)},
    )
    assert resp.status_code == HTTPStatus.OK
    assert (
        b"VRS_Allele_IDs=ga4gh:VA.5PqxTNMJZYJqQZ8MgF_77I1I_qcddGN_,ga4gh:VA._QhHH18HBAIeLos6npRgR-S_0lAX5KR6"
        in resp.content
    )


@pytest.fixture
def sample_vcf_grch37():
    """Basic GRCh37 VCF fixture to use in tests."""
    file_content = b"""##fileformat=VCFv4.2
##hailversion=0.2.100-2ea2615a797a
##INFO=<ID=QUALapprox,Number=1,Type=Integer,Description="">
##INFO=<ID=SB,Number=.,Type=Integer,Description="">
##INFO=<ID=MQ,Number=1,Type=Float,Description="">
##INFO=<ID=MQRankSum,Number=1,Type=Float,Description="">
##INFO=<ID=VarDP,Number=1,Type=Integer,Description="">
##INFO=<ID=AS_ReadPosRankSum,Number=1,Type=Float,Description="">
##INFO=<ID=AS_pab_max,Number=1,Type=Float,Description="">
##INFO=<ID=AS_QD,Number=1,Type=Float,Description="">
##INFO=<ID=AS_MQ,Number=1,Type=Float,Description="">
##INFO=<ID=QD,Number=1,Type=Float,Description="">
##INFO=<ID=AS_MQRankSum,Number=1,Type=Float,Description="">
##INFO=<ID=FS,Number=1,Type=Float,Description="">
##INFO=<ID=AS_FS,Number=1,Type=Float,Description="">
##INFO=<ID=ReadPosRankSum,Number=1,Type=Float,Description="">
##INFO=<ID=AS_QUALapprox,Number=1,Type=Integer,Description="">
##INFO=<ID=AS_SB_TABLE,Number=.,Type=Integer,Description="">
##INFO=<ID=AS_VarDP,Number=1,Type=Integer,Description="">
##INFO=<ID=AS_SOR,Number=1,Type=Float,Description="">
##INFO=<ID=SOR,Number=1,Type=Float,Description="">
##INFO=<ID=singleton,Number=0,Type=Flag,Description="">
##INFO=<ID=transmitted_singleton,Number=0,Type=Flag,Description="">
##INFO=<ID=omni,Number=0,Type=Flag,Description="">
##INFO=<ID=mills,Number=0,Type=Flag,Description="">
##INFO=<ID=monoallelic,Number=0,Type=Flag,Description="">
##INFO=<ID=AS_VQSLOD,Number=1,Type=Float,Description="">
##INFO=<ID=InbreedingCoeff,Number=1,Type=Float,Description="">
##FILTER=<ID=AC0,Description="Allele count is zero after filtering out low-confidence genotypes (GQ < 20; DP < 10; and AB < 0.2 for het calls)">
##FILTER=<ID=AS_VQSR,Description="Failed VQSR filtering thresholds of -2.7739 for SNPs and -1.0606 for indels">
##contig=<ID=11,length=248956422,assembly=GRCh37>
#CHROM	POS	ID	REF	ALT	QUAL	FILTER	INFO
11	47352948	.	TGAGAA	CCT	.	PASS	QUALapprox=21493;SB=325,1077,113,694;MQ=32.1327;MQRankSum=0.720000;VarDP=2236;AS_ReadPosRankSum=-0.736000;AS_pab_max=1.00000;AS_QD=5.17857;AS_MQ=29.5449;QD=9.61225;AS_MQRankSum=0.00000;FS=8.55065;AS_FS=.;ReadPosRankSum=0.727000;AS_QUALapprox=145;AS_SB_TABLE=325,1077,2,5;AS_VarDP=28;AS_SOR=0.311749;SOR=1.48100;singleton;AS_VQSLOD=13.4641;InbreedingCoeff=-0.000517845"""
    return io.BytesIO(file_content)


def test_vcf_registration_grch37(
    restapi_client: TestClient, sample_vcf_grch37: io.BytesIO
):
    """Test registration and annotation of VCFs with GRCh37 assembly"""
    resp = restapi_client.put(
        "/vcf",
        params={"assembly": "GRCh37"},
        files={"vcf": ("test.vcf", sample_vcf_grch37)},
    )
    assert resp.status_code == HTTPStatus.OK
    assert (
        b"VRS_Allele_IDs=ga4gh:VA.STbZFTK0grB7JO2cs3TCblpDiMjIMJWq,ga4gh:VA.CNSRLQlBrly3rRcdldN85dw2Tjos7Cas"
        in resp.content
    )


def test_vcf_registration_invalid_assembly(
    restapi_client: TestClient, sample_vcf_grch37: io.BytesIO
):
    """Test registration and annotation of VCFs with invalid assembly param value"""
    resp = restapi_client.put(
        "/vcf",
        params={"assembly": "hg19"},
        files={"vcf": ("test.vcf", sample_vcf_grch37)},
    )
    assert resp.status_code == HTTPStatus.UNPROCESSABLE_ENTITY


def test_vcf_registration_async(
    restapi_client: TestClient,
    sample_vcf_grch38: io.BytesIO,
    celery_context,  # noqa: ARG001
    vcf_run_id: str,
):
    """Test the async VCF annotation process using a real Celery worker and background task"""
    resp = restapi_client.put(
        "/vcf",
        params={"assembly": "GRCh38", "run_id": vcf_run_id, "run_async": True},
        files={"vcf": ("test.vcf", sample_vcf_grch38)},
    )
    assert resp.status_code == HTTPStatus.ACCEPTED, resp.text
    assert "status_message" in resp.json()
    assert (
        resp.json()["status_message"]
        == f"Run submitted. Check status at /vcf/{vcf_run_id}"
    )
    assert "status" in resp.json()
    assert resp.json()["status"] == "PENDING"
    assert "run_id" in resp.json()
    assert resp.json()["run_id"] == vcf_run_id

    while True:
        resp = restapi_client.get(f"/vcf/{vcf_run_id}")
        if resp.status_code == HTTPStatus.ACCEPTED:
            time.sleep(1)
        elif resp.status_code == HTTPStatus.OK:
            break
        else:
            raise AssertionError(f"Unexpected HTTP response: {resp.status_code}")

    assert (
        b"VRS_Allele_IDs=ga4gh:VA.5PqxTNMJZYJqQZ8MgF_77I1I_qcddGN_,ga4gh:VA._QhHH18HBAIeLos6npRgR-S_0lAX5KR6"
        in resp.content
    )


def test_vcf_submit_no_async(
    restapi_client: TestClient, sample_vcf_grch38: io.BytesIO, mocker: MockerFixture
):
    """Tests that a 400 is returned when async processing is not enabled"""
    mocker.patch.dict(
        os.environ, {"ANYVAR_VCF_ASYNC_WORK_DIR": "", "CELERY_BROKER_URL": ""}
    )
    resp = restapi_client.put(
        "/vcf",
        params={"assembly": "GRCh38", "run_id": "12345", "run_async": True},
        files={"vcf": ("test.vcf", sample_vcf_grch38)},
    )
    assert resp.status_code == HTTPStatus.BAD_REQUEST
    assert "error" in resp.json()
    assert (
        resp.json()["error"]
        == "Required modules and/or configurations for asynchronous VCF annotation are missing"
    )


def test_vcf_submit_duplicate_run_id(
    restapi_client: TestClient, sample_vcf_grch38: io.BytesIO, mocker: MockerFixture
):
    """Tests the submit VCF endpoint when there is already a run for the specified run id"""
    mocker.patch.dict(
        os.environ, {"ANYVAR_VCF_ASYNC_WORK_DIR": "./", "CELERY_BROKER_URL": "redis://"}
    )
    mock_result = mocker.patch("anyvar.restapi.vcf_router.AsyncResult")
    mock_result.return_value.status = "SENT"
    resp = restapi_client.put(
        "/vcf",
        params={"assembly": "GRCh38", "run_id": "12345", "run_async": True},
        files={"vcf": ("test.vcf", sample_vcf_grch38)},
    )
    assert resp.status_code == HTTPStatus.BAD_REQUEST
    assert "error" in resp.json()
    assert (
        resp.json()["error"]
        == "An existing run with id 12345 is SENT. Fetch the completed run result before submitting with the same run_id."
    )


def test_vcf_get_result_no_async(restapi_client: TestClient, mocker: MockerFixture):
    """Tests that a 400 is returned when async processing is not enabled"""
    mocker.patch.dict(
        os.environ, {"ANYVAR_VCF_ASYNC_WORK_DIR": "", "CELERY_BROKER_URL": ""}
    )
    resp = restapi_client.get("/vcf/12345")
    assert resp.status_code == HTTPStatus.BAD_REQUEST
    assert "error" in resp.json()
    assert (
        resp.json()["error"]
        == "Required modules and/or configurations for asynchronous VCF annotation are missing"
    )


def test_vcf_get_result_success(restapi_client: TestClient, mocker: MockerFixture):
    """Tests the get async VCF result endpoint when annotation was successful"""
    mocker.patch.dict(
        os.environ, {"ANYVAR_VCF_ASYNC_WORK_DIR": "./", "CELERY_BROKER_URL": "redis://"}
    )
    mock_result = mocker.patch("anyvar.restapi.vcf_router.AsyncResult")
    mock_result.return_value.status = "SUCCESS"
    mock_result.return_value.result = __file__
    mock_bg_tasks = mocker.patch("anyvar.restapi.vcf_router.BackgroundTasks.add_task")
    resp = restapi_client.get("/vcf/12345")
    assert resp.status_code == HTTPStatus.OK
    with pathlib.Path(__file__).open(mode="rb") as fd:
        assert resp.content == fd.read()
    mock_result.return_value.forget.assert_called_once()
    mock_bg_tasks.assert_called_with(_working_file_cleanup, __file__)


def test_vcf_get_result_failure_timeout(
    restapi_client: TestClient, mocker: MockerFixture
):
    """Tests the get async VCF result endpoint when annotation fails due to task timeout"""
    mocker.patch.dict(
        os.environ, {"ANYVAR_VCF_ASYNC_WORK_DIR": "./", "CELERY_BROKER_URL": "redis://"}
    )
    mock_result = mocker.patch("anyvar.restapi.vcf_router.AsyncResult")
    mock_result.return_value.status = "FAILURE"
    mock_result.return_value.result = TimeLimitExceeded("task timed out")
    mock_result.return_value.kwargs = {"input_file_path": __file__}
    mock_bg_tasks = mocker.patch("anyvar.restapi.vcf_router.BackgroundTasks.add_task")
    resp = restapi_client.get("/vcf/12345")
    assert resp.status_code == HTTPStatus.INTERNAL_SERVER_ERROR
    assert "error" in resp.json()
    assert resp.json()["error"] == "TimeLimitExceeded('task timed out',)"
    assert "error_code" in resp.json()
    assert resp.json()["error_code"] == "TIME_LIMIT_EXCEEDED"
    mock_result.return_value.forget.assert_called_once()
    mock_bg_tasks.assert_called_once()


def test_vcf_get_result_failure_worker_lost(
    restapi_client: TestClient, mocker: MockerFixture
):
    """Tests the get async VCF result endpoint when annotation failed due to lost worker"""
    mocker.patch.dict(
        os.environ, {"ANYVAR_VCF_ASYNC_WORK_DIR": "./", "CELERY_BROKER_URL": "redis://"}
    )
    mock_result = mocker.patch("anyvar.restapi.vcf_router.AsyncResult")
    mock_result.return_value.status = "FAILURE"
    mock_result.return_value.result = WorkerLostError("killed")
    mock_result.return_value.kwargs = {"input_file_path": __file__}
    mock_bg_tasks = mocker.patch("anyvar.restapi.vcf_router.BackgroundTasks.add_task")
    resp = restapi_client.get("/vcf/12345")
    assert resp.status_code == HTTPStatus.INTERNAL_SERVER_ERROR
    assert "error" in resp.json()
    assert resp.json()["error"] == "killed"
    assert "error_code" in resp.json()
    assert resp.json()["error_code"] == "WORKER_LOST_ERROR"
    mock_result.return_value.forget.assert_called_once()
    mock_bg_tasks.assert_called_once()


def test_vcf_get_result_failure_other(
    restapi_client: TestClient, mocker: MockerFixture
):
    """Tests the get async VCF result endpoint when annotation failed due to an error"""
    mocker.patch.dict(
        os.environ,
        {
            "ANYVAR_VCF_ASYNC_WORK_DIR": "./",
            "CELERY_BROKER_URL": "redis://",
            "ANYVAR_VCF_ASYNC_FAILURE_STATUS_CODE": "200",
        },
    )
    mock_result = mocker.patch("anyvar.restapi.vcf_router.AsyncResult")
    mock_result.return_value.status = "FAILURE"
    mock_result.return_value.result = KeyError("foo")
    mock_result.return_value.kwargs = {"input_file_path": __file__}
    mock_bg_tasks = mocker.patch("anyvar.restapi.vcf_router.BackgroundTasks.add_task")
    resp = restapi_client.get("/vcf/12345")
    assert resp.status_code == HTTPStatus.OK
    assert "error" in resp.json()
    assert resp.json()["error"] == "'foo'"
    assert "error_code" in resp.json()
    assert resp.json()["error_code"] == "RUN_FAILURE"
    mock_result.return_value.forget.assert_called_once()
    mock_bg_tasks.assert_called_once()


def test_vcf_get_result_notfound(restapi_client: TestClient, mocker: MockerFixture):
    """Tests the get async VCF result endpoint when run id is not found"""
    mocker.patch.dict(
        os.environ, {"ANYVAR_VCF_ASYNC_WORK_DIR": "./", "CELERY_BROKER_URL": "redis://"}
    )
    mock_result = mocker.patch("anyvar.restapi.vcf_router.AsyncResult")
    mock_result.return_value.status = "PENDING"
    resp = restapi_client.get("/vcf/12345")
    assert resp.status_code == HTTPStatus.NOT_FOUND
    assert "status_message" in resp.json()
    assert resp.json()["status_message"] == "Run not found"
    assert "status" in resp.json()
    assert resp.json()["status"] == "NOT_FOUND"
    assert "run_id" in resp.json()
    assert resp.json()["run_id"] == "12345"


def test_vcf_get_result_notcomplete(restapi_client: TestClient, mocker: MockerFixture):
    """Tests the get async VCF result endpoint when annotation is not yet complete"""
    mocker.patch.dict(
        os.environ, {"ANYVAR_VCF_ASYNC_WORK_DIR": "./", "CELERY_BROKER_URL": "redis://"}
    )
    mock_result = mocker.patch("anyvar.restapi.vcf_router.AsyncResult")
    mock_result.return_value.status = "SENT"
    resp = restapi_client.get("/vcf/12345")
    assert resp.status_code == HTTPStatus.ACCEPTED
    assert "status_message" in resp.json()
    assert (
        resp.json()["status_message"] == "Run not completed. Check status at /vcf/12345"
    )
    assert "status" in resp.json()
    assert resp.json()["status"] == "PENDING"
    assert "run_id" in resp.json()
    assert resp.json()["run_id"] == "12345"



"""Provide response definitions to REST API endpoint."""

from collections.abc import Iterable
from datetime import datetime
from enum import Enum, StrEnum
from typing import Any

from ga4gh.vrs import (
    VRS_VERSION,
    VrsType,
    models,
)
from ga4gh.vrs import (
    __version__ as vrs_python_version,
)
from pydantic import BaseModel, ConfigDict, Field, StrictStr

from anyvar import __version__
from anyvar.core import metadata, objects
from anyvar.mapping.liftover import ReferenceAssembly


class EndpointTag(str, Enum):
    """Denote endpoint group membership"""

    META = "Service Metadata"
    VCF = "VCF Operations"
    VRS_OBJECTS = "VRS Objects"
    SEARCH = "Search Operations"


class ServiceEnvironment(str, Enum):
    """Define current runtime environment."""

    DEV = "dev"
    PROD = "prod"
    TEST = "test"
    STAGING = "staging"


class ServiceOrganization(BaseModel):
    """Define service_info response for organization field"""

    name: str = Field(
        default="bioccommons",
        description="Name of the organization responsible for the service",
    )
    url: str = Field(
        default="https://biocommons.org",
        description="URL of the website of the organization (RFC 3986 format)",
    )


class ServiceType(BaseModel):
    """Define service_info response for type field"""

    group: str = Field(
        default="org.biocommons",
        description="Namespace in reverse domain name format. Use `org.ga4gh` for implementations compliant with official GA4GH specifications. For services with custom APIs not standardized by GA4GH, or implementations diverging from official GA4GH specifications, use a different namespace (e.g. your organization''s reverse domain name).",
    )
    artifact: str = Field(
        default="anyvar",
        description="Name of the API or GA4GH specification implemented. Official GA4GH types should be assigned as part of standards approval process. Custom artifacts are supported.",
    )
    version: str = __version__


class SpecMetadata(BaseModel):
    """Define substructure for reporting specification metadata."""

    vrs_version: str = VRS_VERSION


class ImplMetadata(BaseModel):
    """Define substructure for reporting metadata about internal software dependencies."""

    vrs_python_version: str = vrs_python_version


class CapabilitiesMetadata(BaseModel):
    """Define substructure for reporting metadata about service capabilities"""

    liftover_assemblies: list[str] = ["GRCh38", "GRCh37"]


class ServiceInfo(BaseModel):
    """Define response structure for GA4GH /service_info endpoint."""

    id: str = Field(
        default="org.biocommons.anyvar",
        description="Unique ID of this service. Reverse domain name notation is recommended, though not required. The identifier should attempt to be globally unique so it can be used in downstream aggregator services e.g. Service Registry.",
    )
    name: str = Field(
        default="AnyVar", description="Name of this service. Should be human readable."
    )
    type: ServiceType = Field(
        default=ServiceType(), description="Type of a GA4GH service"
    )
    description: str = Field(
        default="Register and retrieve variations and associated annotations.",
        description="Description of the service. Should be human readable and provide information about the service.",
    )
    organization: ServiceOrganization = Field(
        default=ServiceOrganization(), description="Organization providing the service"
    )
    contactUrl: str = Field(  # noqa: N815
        default="mailto:alex.wagner@nationwidechildrens.org",
        description="URL of the contact for the provider of this service, e.g. a link to a contact form (RFC 3986 format), or an email (RFC 2368 format).",
    )
    documentationUrl: str = Field(  # noqa: N815
        default="https://github.com/biocommons/anyvar",
        description="URL of the documentation of this service (RFC 3986 format). This should help someone learn how to use your service, including any specifics required to access data, e.g. authentication.",
    )
    createdAt: datetime = Field(  # noqa: N815
        default=datetime.fromisoformat("2025-06-01T00:00:00Z"),
        description="Timestamp describing when the service was first deployed and available (RFC 3339 format)",
    )
    updatedAt: datetime = Field(  # noqa: N815
        default=datetime.fromisoformat("2025-06-01T00:00:00Z"),
        description="Timestamp describing when the service was last updated (RFC 3339 format)",
    )
    environment: ServiceEnvironment = Field(
        default=ServiceEnvironment.DEV,
        description="Environment the service is running in. Use this to distinguish between production, development and testing/staging deployments. Suggested values are prod, test, dev, staging. However this is advised and not enforced.",
    )
    version: str = __version__
    spec_metadata: SpecMetadata = SpecMetadata()
    impl_metadata: ImplMetadata = ImplMetadata()
    capabilities_metadata: CapabilitiesMetadata = CapabilitiesMetadata()


class TranslationResult(BaseModel):
    """Describe structure for translation success or failure"""

    variation: objects.VrsVariation | None = None
    error: str | None = None


class SupportedVariationType(StrEnum):
    """Supported variation types for API input."""

    ALLELE = VrsType.ALLELE.value
    COPY_NUMBER_CHANGE = VrsType.CN_CHANGE.value
    COPY_NUMBER_COUNT = VrsType.CN_COUNT.value


class VariationRequest(BaseModel):
    """Describe request structure for the PUT and POST /variation endpoints"""

    model_config = ConfigDict(use_enum_values=True)

    definition: StrictStr
    input_type: SupportedVariationType | None = None
    copies: int | None = None
    copy_change: models.CopyChange | None = None
    assembly_name: ReferenceAssembly | None = ReferenceAssembly.GRCH38


class AddAnnotationResponse(BaseModel):
    """Response for the POST /variation/{vrs_id}/annotations endpoint"""

    object: objects.VrsObject | None
    object_id: str | None
    annotation_type: str | None
    annotation_value: Any | None
    annotation_id: int | None


class AddAnnotationRequest(BaseModel):
    """Request for the POST /variation/{vrs_id}/annotations endpoint.

    Used when the annotation is identified through the request path.
    """

    annotation_type: str
    annotation_value: Any


class GetAnnotationResponse(BaseModel):
    """Response for the GET /variation/{vrs_id}/annotations/{annotation_type} endpoint"""

    annotations: list[metadata.Annotation]


class AddMappingResponse(BaseModel):
    """Response for POST /variation/{vrs_id}/mappings endpoint"""

    model_config = ConfigDict(use_enum_values=True)

    source_object: objects.VrsObject | None
    source_object_id: str
    dest_object: objects.VrsObject | None
    dest_object_id: str
    mapping_type: metadata.VariationMappingType


class AddMappingRequest(BaseModel):
    """Request for the POST /variation/{vrs_id}/mappings endpoint"""

    model_config = ConfigDict(use_enum_values=True)

    dest_id: str
    mapping_type: metadata.VariationMappingType


class GetMappingResponse(BaseModel):
    """Request for the GET /variation/{vrs_id}/mappings endpoint"""

    mappings: Iterable[metadata.VariationMapping]


class RegisterVariationResponse(BaseModel):
    """Describe response for the PUT /variation, PUT /variations, and PUT /vrs_variation endpoints"""

    model_config = ConfigDict(
        json_schema_extra={
            "examples": [
                {
                    "input_variation": {
                        "definition": "NC_000007.13:g.36561662_36561663del",
                        "input_type": "Allele",
                        "copies": 0,
                        "copy_change": "complete genomic loss",
                        "assembly_name": None,
                    },
                    "messages": [],
                    "object": {
                        "id": "ga4gh:VA.d6ru7RcuVO0-v3TtPFX5fZz-GLQDhMVb",
                        "type": "Allele",
                        "digest": "d6ru7RcuVO0-v3TtPFX5fZz-GLQDhMVb",
                        "location": {
                            "id": "ga4gh:SL.JOFKL4nL5mRUlO_xLwQ8VOD1v7mxhs3I",
                            "type": "SequenceLocation",
                            "digest": "JOFKL4nL5mRUlO_xLwQ8VOD1v7mxhs3I",
                            "sequenceReference": {
                                "type": "SequenceReference",
                                "refgetAccession": "SQ.IW78mgV5Cqf6M24hy52hPjyyo5tCCd86",
                            },
                            "start": 36561661,
                            "end": 36561663,
                        },
                        "state": {
                            "type": "ReferenceLengthExpression",
                            "length": 0,
                            "sequence": "",
                            "repeatSubunitLength": 2,
                        },
                    },
                    "object_id": "ga4gh:VA.d6ru7RcuVO0-v3TtPFX5fZz-GLQDhMVb",
                }
            ]
        }
    )

    input_variation: VariationRequest | objects.VrsVariation
    messages: list[str] = []
    object: objects.VrsVariation | None = None
    object_id: str | None = None


class GetObjectResponse(BaseModel):
    """Describe response for the GET /variation endpoint"""

    model_config = ConfigDict(
        json_schema_extra={
            "examples": [
                {
                    "messages": [],
                    "data": {
                        "digest": "K7akyz9PHB0wg8wBNVlWAAdvMbJUJJfU",
                        "id": "ga4gh:VA.K7akyz9PHB0wg8wBNVlWAAdvMbJUJJfU",
                        "location": {
                            "digest": "01EH5o6V6VEyNUq68gpeTwKE7xOo-WAy",
                            "id": "ga4gh:SL.01EH5o6V6VEyNUq68gpeTwKE7xOo-WAy",
                            "start": 87894076,
                            "end": 87894077,
                            "sequenceReference": {
                                "refgetAccession": "SQ.ss8r_wB0-b9r44TQTMmVTI92884QvBiB",
                                "type": "SequenceReference",
                            },
                            "type": "SequenceLocation",
                        },
                        "state": {"sequence": "T", "type": "LiteralSequenceExpression"},
                        "type": "Allele",
                    },
                }
            ]
        }
    )

    messages: list[StrictStr]
    data: objects.VrsObject | None = None


class SearchResponse(BaseModel):
    """Describe response for the GET /search endpoint"""

    variations: list[models.Variation]


class RunStatusResponse(BaseModel):
    """Represents the response for triggering or checking the status of a run
    at the GET /vcf/{run_id} endpoint.
    """

    run_id: str  # Run ID
    status: str  # Run status
    status_message: str | None = None  # Detailed status message for failures


class ErrorResponse(BaseModel):
    """Represents an error message"""

    error: str  # Error message
    error_code: str | None = None  # error code



import pytest
from ga4gh.vrs import models

from anyvar.anyvar import AnyVar
from anyvar.core.metadata import VariationMappingType
from anyvar.mapping import liftover
from anyvar.storage.base import Storage
from anyvar.translate.base import Translator


# Success Cases
@pytest.fixture(scope="session")
def copynumber_ranged_positive_grch37_variant(copy_number_variations: dict):
    return {
        "grch37": models.CopyNumberCount(
            **copy_number_variations["ga4gh:CN.CTCgVehH0FEqrlaOMhUsDjKwzavnQegk"][
                "variation"
            ]
        ),
        "grch38": models.CopyNumberCount(
            **copy_number_variations["ga4gh:CN.fmrn873tRhAiNLybjHlftgHjcAEExPKQ"][
                "variation"
            ]
        ),
    }


@pytest.fixture(scope="session")
def allele_int_negative_grch38_variant(alleles: dict):
    return {
        "grch37": models.Allele(
            **alleles["ga4gh:VA.nmp-bzYpO00NYIqr3CaVF0ZH2ZpSj1ly"]["variation"]
        ),
        "grch38": models.Allele(
            **alleles["ga4gh:VA.Otc5ovrw906Ack087o1fhegB4jDRqCAe"]["variation"]
        ),
    }


@pytest.fixture
def allele_int_unknown_grch38_variant(alleles: dict):
    return {
        "grch37": models.Allele(
            **alleles["ga4gh:VA.J-gW7La8EblIdT1MfqZzhzbO26lkEH7D"]["variation"]
        ),
        "grch38": models.Allele(
            **alleles["ga4gh:VA.9gW_iJbQAIO3SIxJ9ACyAZA1X2lEgO39"]["variation"]
        ),
    }


@pytest.fixture
def allele_int_rle_grch37_variant(alleles: dict):
    return {
        "grch37": models.Allele(
            **alleles["ga4gh:VA.d6ru7RcuVO0-v3TtPFX5fZz-GLQDhMVb"]["variation"]
        ),
        "grch38": models.Allele(
            **alleles["ga4gh:VA.QiKnpRR8S7SPoUE-RMUJJbT-RS1akuHA"]["variation"]
        ),
    }


# Failure Cases
@pytest.fixture
def grch36_variant(alleles: dict):
    return {
        "variant": models.Allele(
            **alleles["ga4gh:VA.4dEsVNR2JC_ZiHsYSGZgariIUOfYl6a0"]["variation"]
        ),
        "error": liftover.UnsupportedReferenceAssemblyError(
            "Unable to get reference sequence ID for SQ.JY7UegcaYT-M0PYn1yDGQ_4XJsa-DsXq",
        ),
    }


@pytest.fixture
def unconvertible_grch37_variant(alleles: dict):
    return {
        "variant": models.Allele(
            **alleles["ga4gh:VA.qP-qtMJqKhTEJfpTdAZN9CoIFCRKv4kg"]["variation"]
        ),
        "error": liftover.CoordinateConversionFailureError(),
    }


@pytest.fixture
def unconvertible_grch38_variant(alleles: dict):
    return {
        "variant": models.Allele(
            **alleles["ga4gh:VA.5-m9wM6WTY5osPxLFg1_bITsOwSoMFui"]["variation"]
        ),
        "error": liftover.CoordinateConversionFailureError(),
    }


# Cases where liftover should not be attempted
@pytest.fixture
def empty_variation_object():
    return {
        "variant": {},
        "error": None,
    }


@pytest.fixture
def invalid_variant():
    return {
        "variant": {
            "location": {
                "id": "ga4gh:SL.aCMcqLGKClwMWEDx3QWe4XSiGDlKXdB8",
                "end": 0,
                "start": -1,
                "sequenceReference": {
                    "refgetAccession": "SQ.ss8r_wB0-b9r44TQTMmVTI92884QvBiB",
                    "type": "SequenceReference",
                },
                "type": "SequenceLocation",
            },
            "state": {"sequence": "T", "type": "LiteralSequenceExpression"},
            "type": "UNSUPPORTED",
        },
        "error": None,
    }


# Cases where liftover should be successful
SUCCESS_CASES = [
    "allele_int_rle_grch37_variant",
    "allele_int_negative_grch38_variant",
    "allele_int_unknown_grch38_variant",
]

# Cases where liftover should be unsuccessful
FAILURE_CASES = [
    "grch36_variant",
    "unconvertible_grch37_variant",
    "unconvertible_grch38_variant",
]

# Cases where liftover should not be attempted
NO_LIFTOVER_CASES = ["empty_variation_object", "invalid_variant"]


####################################################################################
## Tests for src/anyvar/mapping/liftover.py > 'get_liftover_variant' function ##
####################################################################################
@pytest.mark.parametrize("variant_fixture_name", SUCCESS_CASES)
@pytest.mark.ci_ok
def test_liftover_success(request: pytest.FixtureRequest, variant_fixture_name: str):
    fixture = request.getfixturevalue(variant_fixture_name)
    grch37 = fixture["grch37"]
    grch38 = fixture["grch38"]

    # 37 to 38
    lifted_over_variant_output = liftover.get_liftover_variant(grch37)
    assert lifted_over_variant_output == grch38

    # 38 to 37
    lifted_over_variant_output = liftover.get_liftover_variant(grch38)
    assert lifted_over_variant_output == grch37


@pytest.mark.parametrize("variant_fixture_name", FAILURE_CASES)
@pytest.mark.ci_ok
def test_liftover_failure(request, variant_fixture_name):
    fixture = request.getfixturevalue(variant_fixture_name)
    variant_input = fixture["variant"]
    expected_error = fixture["error"]
    with pytest.raises(
        type(expected_error),
        match=expected_error.args[0] if expected_error.args else None,
    ):
        liftover.get_liftover_variant(variant_input)


######################################################################################################
## Tests for `src/anyvar/mapping/liftover.py > 'add_liftover_mapping' ##
######################################################################################################
@pytest.mark.parametrize(
    "variant_fixture_name",
    SUCCESS_CASES,
)
@pytest.mark.parametrize(
    ("src_key", "dst_key"),
    [("grch37", "grch38"), ("grch38", "grch37")],
    ids=["37->38", "38->37"],
)
def test_liftover_mapping_success(
    request,
    variant_fixture_name,
    src_key: str,
    dst_key: str,
    storage: Storage,
    translator: Translator,
):
    fixture = request.getfixturevalue(variant_fixture_name)
    src = fixture[src_key]
    dst = fixture[dst_key]

    # ensure input is present in DB
    storage.add_objects([src])
    liftover.add_liftover_mapping(src, storage, translator.dp)

    # mapping and lifted-over variant should be present
    mappings = list(storage.get_mappings(src.id, VariationMappingType.LIFTOVER))
    assert len(mappings) == 1
    assert mappings[0].dest_id == dst.id

    result = list(storage.get_objects(models.Allele, [dst.id]))
    assert len(result) == 1


@pytest.mark.parametrize(
    "variant_fixture_name",
    FAILURE_CASES,
)
def test_liftover_mapping_failure(
    request,
    variant_fixture_name,
    anyvar_instance: AnyVar,
    storage: Storage,
    translator: Translator,
):
    fixture = request.getfixturevalue(variant_fixture_name)
    variant_input = fixture["variant"]
    expected_error = fixture["error"]

    # ensure input is present in DB
    anyvar_instance.object_store.add_objects([variant_input])

    assert liftover.add_liftover_mapping(variant_input, storage, translator.dp) == [
        expected_error.get_error_message()
    ]



"""Provide tools for and implementations of AnyVar storage backends."""

from .base import Storage

DEFAULT_STORAGE_URI = "postgresql://postgres@localhost:5432/anyvar"

__all__ = ["DEFAULT_STORAGE_URI", "Storage"]



"""Tests Snowflake storage operations with dynamic tables enabled.
Only the functionality specific to the dynamic tables feature is tested here."""

import os

import pytest
from ga4gh.vrs import models
from sqlalchemy import text

from anyvar.storage import orm
from anyvar.storage.snowflake import SnowflakeObjectStore

pytestmark = pytest.mark.snowflake


@pytest.fixture(scope="module")
def monkeymodule():
    """A module-scoped monkeypatch fixture."""
    with pytest.MonkeyPatch.context() as mp:
        yield mp


@pytest.fixture(scope="module")
def snowflake_storage_with_dyn_tables(monkeymodule):
    # first create the storage connector and drop any existing db state
    storage = SnowflakeObjectStore(db_url=os.environ.get("ANYVAR_TEST_STORAGE_URI"))
    with storage.engine.connect() as conn:
        conn.execute(text(f"DROP TABLE IF EXISTS {orm.Annotation.__tablename__}"))
        conn.execute(text(f"DROP TABLE IF EXISTS {orm.VariationMapping.__tablename__}"))
        conn.execute(text(f"DROP TABLE IF EXISTS {orm.Allele.__tablename__}"))
        conn.execute(text(f"DROP TABLE IF EXISTS {orm.Location.__tablename__}"))
        conn.execute(
            text(f"DROP TABLE IF EXISTS {orm.SequenceReference.__tablename__}")
        )
        conn.execute(text(f"DROP TABLE IF EXISTS {orm.VrsObject.__tablename__}"))

    # then re-create the storage connector with dynamic tables enabled
    monkeymodule.setenv("ANYVAR_SNOWFLAKE_STORE_USE_DYNAMIC_TABLES", "true")
    monkeymodule.setenv("ANYVAR_SNOWFLAKE_STORE_USE_JOIN_FOR_MERGE", "false")
    storage = SnowflakeObjectStore(db_url=os.environ.get("ANYVAR_TEST_STORAGE_URI"))
    yield storage

    # and finally, drop the tables again
    with storage.engine.connect() as conn:
        conn.execute(text(f"DROP TABLE IF EXISTS {orm.Annotation.__tablename__}"))
        conn.execute(text(f"DROP TABLE IF EXISTS {orm.VariationMapping.__tablename__}"))
        conn.execute(text(f"DROP TABLE IF EXISTS {orm.Allele.__tablename__}"))
        conn.execute(text(f"DROP TABLE IF EXISTS {orm.Location.__tablename__}"))
        conn.execute(
            text(f"DROP TABLE IF EXISTS {orm.SequenceReference.__tablename__}")
        )
        conn.execute(text(f"DROP TABLE IF EXISTS {orm.VrsObject.__tablename__}"))


def test_add_alleles(
    snowflake_storage_with_dyn_tables: SnowflakeObjectStore,
    focus_alleles: tuple[models.Allele, models.Allele, models.Allele],
):
    """Test adding alleles to snowflake storage with dynamic tables enabled."""

    # add the alleles
    snowflake_storage_with_dyn_tables.add_objects(focus_alleles)

    # refresh the dynamic tables to ensure data is available for querying
    with snowflake_storage_with_dyn_tables.engine.connect() as conn:
        conn.execute(text(f"ALTER DYNAMIC TABLE {orm.Allele.__tablename__} REFRESH"))
        conn.execute(text(f"ALTER DYNAMIC TABLE {orm.Location.__tablename__} REFRESH"))
        conn.execute(
            text(f"ALTER DYNAMIC TABLE {orm.SequenceReference.__tablename__} REFRESH")
        )

    # get 1 allele
    result = snowflake_storage_with_dyn_tables.get_objects(
        models.Allele, [focus_alleles[0].id]
    )
    assert result == [focus_alleles[0]]

    # get 2 sequence locations
    result = snowflake_storage_with_dyn_tables.get_objects(
        models.SequenceLocation,
        [focus_alleles[1].location.id, focus_alleles[0].location.id],
    )
    assert len(result) == 2
    assert focus_alleles[0].location in result
    assert focus_alleles[1].location in result

    # get a sequence reference
    result = snowflake_storage_with_dyn_tables.get_objects(
        models.SequenceReference,
        [
            "ga4gh:VA.1FzYrqG-7jB3Wr46eIL_L5BWElQZEB7i",
            focus_alleles[0].location.sequenceReference.refgetAccession,
        ],
    )
    assert result == [focus_alleles[0].location.sequenceReference]

    # delete data
    snowflake_storage_with_dyn_tables.wipe_db()

    # refresh the dynamic tables to ensure data is available for querying
    with snowflake_storage_with_dyn_tables.engine.connect() as conn:
        conn.execute(text(f"ALTER DYNAMIC TABLE {orm.Allele.__tablename__} REFRESH"))
        conn.execute(text(f"ALTER DYNAMIC TABLE {orm.Location.__tablename__} REFRESH"))
        conn.execute(
            text(f"ALTER DYNAMIC TABLE {orm.SequenceReference.__tablename__} REFRESH")
        )

    # verify that the dynamic tables are empty after wipe
    result = snowflake_storage_with_dyn_tables.get_objects(
        models.Allele, [focus_alleles[0].id]
    )
    assert result == []

    # get 2 sequence locations
    result = snowflake_storage_with_dyn_tables.get_objects(
        models.SequenceLocation,
        [focus_alleles[1].location.id, focus_alleles[0].location.id],
    )
    assert result == []

    # get a sequence reference
    result = snowflake_storage_with_dyn_tables.get_objects(
        models.SequenceReference,
        [
            focus_alleles[0].location.sequenceReference.refgetAccession,
        ],
    )
    assert result == []



"""Defines functions used to lift over variants between GRCh37 & GRCh38"""

import logging
from enum import Enum
from typing import TypeVar

from agct import (
    Assembly,
    Converter,
    Strand,
    get_converter,
    get_refget_id_from_seqinfo,
    get_seqinfo_from_refget_id,
)
from ga4gh.core import ga4gh_identify
from ga4gh.vrs import models, normalize
from ga4gh.vrs.dataproxy import _DataProxy

from anyvar.core.metadata import VariationMapping, VariationMappingType
from anyvar.core.objects import VrsVariation
from anyvar.storage.base import Storage

_logger = logging.getLogger(__name__)


class ReferenceAssembly(Enum):
    """Supported reference assemblies"""

    GRCH37 = "GRCh37"
    GRCH38 = "GRCh38"


class LiftoverError(Exception):
    """Indicates a failure to liftover a variant between GRCh37 & GRCh38"""

    base_error_message = "Unable to complete liftover"
    error_details = ""

    @classmethod
    def get_error_message(cls) -> str:
        """Return the error message associated with the Exception"""
        return (
            f"{cls.base_error_message}: {cls.error_details}"
            if cls.error_details
            else cls.base_error_message
        )


class UnsupportedVariantLocationTypeError(LiftoverError):
    """Indicates a variant with a 'location' type that is unsupported"""

    error_details = "Liftover is unsupported for variants without refget accession, start position, and end position"


class UnsupportedReferenceAssemblyError(LiftoverError):
    """Indicates a failure to retrieve alias data for a refget accession in any supported reference assembly."""

    error_details = "Could not resolve reference assembly - accession not found in any supported assembly"


class CoordinateConversionFailureError(LiftoverError):
    """Indicates a failure to lift over a variant's coordinate"""

    error_details = "Could not convert start and/or end position(s)"


class AmbiguousCoordinateConversionError(LiftoverError):
    """Indicates AnyVar cannot lift over a variant because its start and/or end coordinates mapped to multiple possible locations"""

    error_details = "Start and/or end positions mapped to multiple possible locations"


class AccessionConversionError(LiftoverError):
    """Indicates a failure to convert a variant's refget accession"""

    error_details = "Could not convert refget accession"


def _convert_coordinate(converter: Converter, chromosome: str, coordinate: int) -> int:
    """Convert an individual coordinate to another reference genome. If the conversion is unsuccessful, raises a `CoordinateConversionError`

    :param converter: An AGCT Converter instance.
    :param chromosome: The chromosome number where the coordinate is found. Must be a string consisting of a) the prefix "chr", and b) a number OR "X" or "Y" -> i.e. "chr10".
    :param coordinate: A single start or end coordinate. MUST be an `int`.
    :return: A converted coordinate value as an `int`
    :raises: A `CoordinateConversionFailureError` if the conversion returns no results.
    :raises: A `AmbiguousCoordinateConversionError` if the conversion returns more than one result.
    """
    converted_positions = converter.convert_coordinate(
        chromosome, coordinate, coordinate, Strand.POSITIVE
    )  # returns a list of tuples with a) the coordinate's chromosome number, b) the converted coordinate position, and c) the strand (Strand.POSITIVE or Strand.NEGATIVE)

    if len(converted_positions) > 1:
        raise AmbiguousCoordinateConversionError

    if (
        len(converted_positions) == 1
        and converted_positions[0].strand
        == Strand.POSITIVE  # TODO: Handle cases where coordinate converts to the negative strand. See Issue #197.
    ):
        return converted_positions[0].start

    raise CoordinateConversionFailureError


_PositionType = TypeVar("_PositionType", int, models.Range)


def convert_position(
    converter: Converter, chromosome: str, position: _PositionType
) -> _PositionType:
    """Convert a SequenceLocation position (i.e., `start` or `end`) to another reference Genome. `position` can either be a `models.Range` or an `int` - return type will match.

    :param converter: An AGCT Converter instance.
    :param chromosome: The chromosome number where the position is found. Must be a string consisting of a) the prefix "chr", and b) a number OR "X" or "Y" -> e.g. "chr10", "chrX", etc.
    :param position: A SequenceLocation start or end position. Can be a `models.Range` or an `int`.

    :return: A lifted-over position. Type (`models.Range` or `int`) will match that of `position`
    """
    # Handle int positions
    if isinstance(position, int):
        return _convert_coordinate(converter, chromosome, position)

    # Handle Range positions
    lower_bound, upper_bound = position.root
    lower_bound = (
        _convert_coordinate(converter, chromosome, lower_bound) if lower_bound else None
    )
    upper_bound = (
        _convert_coordinate(converter, chromosome, upper_bound) if upper_bound else None
    )
    return models.Range([lower_bound, upper_bound])


def get_liftover_variant(input_variant: VrsVariation) -> VrsVariation:
    """Liftover a variant from GRCh37 or GRCH38 into the opposite assembly, and return the converted variant as a VrsVariation.

    If liftover is unsuccessful, raise an Exception.

    :param input_variant: A `VrsVariation`.
    :return: The converted variant as a `VrsVariation`.
    :raises UnsupportedVariantLocationTypeError: If the variant lacks a refget accession, start position or end position
    :raises UnsupportedReferenceAssemblyError: If the variant's accession was not found in any supported assembly
    :raises CoordinateConversionFailureError: If unable to lift over the variant's start and/or end position(s)
    :raises AmbiguousCoordinateConversionError: If variant's start and/or end position(s) map to multiple possible locations
    :raises AccessionConversionError: If unable to lift over the variant's refget accession
    """
    try:
        refget_accession = input_variant.location.sequenceReference.refgetAccession
        start_position = input_variant.location.start
        end_position = input_variant.location.end
    except AttributeError as err:
        raise UnsupportedVariantLocationTypeError from err

    seqinfo = get_seqinfo_from_refget_id(refget_accession)
    if not seqinfo:
        msg = f"Unable to get reference sequence ID for {refget_accession}"
        _logger.error(msg)
        raise UnsupportedReferenceAssemblyError(msg)
    assembly, chromosome = seqinfo
    to_assembly = Assembly.HG19 if assembly == Assembly.HG38 else Assembly.HG38
    converter = get_converter(assembly, to_assembly)

    # Get converted start/end positions. `convert_position` will raise a `CoordinateConversionError` if unsuccessful
    converted_start = convert_position(converter, chromosome, start_position)  # type: ignore (`converter` and `start_position` will always be valid)
    converted_end = convert_position(converter, chromosome, end_position)  # type: ignore (`converter` and `end_position` will always be valid)

    # Get converted refget_accession (without 'ga4gh:' prefix)
    new_alias = f"{to_assembly}:{chromosome}"
    converted_refget_accession = get_refget_id_from_seqinfo(to_assembly, chromosome)
    if not converted_refget_accession:
        # should be impossible
        _logger.error(
            "Unable to convert constructed sequence ID `%s` into refgetAccession",
            new_alias,
        )
        raise AccessionConversionError

    # Build the converted location object
    converted_variant_location = models.SequenceLocation(
        start=converted_start,
        end=converted_end,
        type="SequenceLocation",
        sequenceReference=models.SequenceReference(
            type="SequenceReference", refgetAccession=converted_refget_accession
        ),
    )
    ga4gh_identify(converted_variant_location, in_place="always")

    # Build the liftover variant object
    # Start by copying the original variant
    converted_variant = input_variant.model_copy(deep=True)

    # Replace the location with the lifted-over version
    converted_variant.location = converted_variant_location

    # Recompute object ID
    converted_variant.digest = None
    converted_variant.id = ga4gh_identify(converted_variant, in_place="always")

    return converted_variant


def add_liftover_mapping(
    variation: VrsVariation, storage: Storage, dataproxy: _DataProxy
) -> list[str] | None:
    """Perform liftover between GRCh37 <-> GRCh38. Store mappings between the original and lifted-over variants.

    Don't register lifted-over variant or mappings if
    * liftover fails in either direction
    * liftover is ambiguous in either direction
    * liftover fails to roundtrip accurately

    This function is intended to be quite directly 'user-facing', i.e. it catches and
    suppresses major error cases and communicates results as they are to be transmitted
    in the REST API routes. Library users hoping for more direct control will want to
    employ ``get_liftover_variant`` to perform liftover and make their own decisions
    about when to register objects and mappings.

    :param variation: variation to attempt liftover upon
    :param storage: Storage instance
    :param dataproxy: SeqRepo DataProxy instance, for normalizing lifted-over alleles
    :return: list of messages describing warnings or failures, or ``None`` if completely successful
    """
    input_vrs_id: str = variation.id  # type: ignore
    try:
        lifted_over_variant = get_liftover_variant(variation)
        reverse_liftover_variant = get_liftover_variant(lifted_over_variant)
    except LiftoverError as e:
        _logger.exception(
            "Encountered error during liftover of variation `%s`",
            variation,
        )
        return [e.get_error_message()]

    if reverse_liftover_variant.id != variation.id:
        return [
            f"{LiftoverError.base_error_message}: Roundtripped lifted-over id `{reverse_liftover_variant.id}` does not match initial value of {input_vrs_id}"
        ]

    normalized_lifted_over_variant = normalize(
        lifted_over_variant, data_proxy=dataproxy
    )
    normalized_lifted_over_variant_id: str = normalized_lifted_over_variant.id  # type: ignore
    storage.add_objects([normalized_lifted_over_variant])
    storage.add_mapping(
        VariationMapping(
            source_id=input_vrs_id,
            dest_id=normalized_lifted_over_variant_id,
            mapping_type=VariationMappingType.LIFTOVER,
        )
    )
    storage.add_mapping(
        VariationMapping(
            source_id=normalized_lifted_over_variant_id,
            dest_id=input_vrs_id,
            mapping_type=VariationMappingType.LIFTOVER,
        )
    )
    return None
